# -*- encoding: utf-8 -*-
"""Defines a set of mixins that provide tools for interacting with entities."""
import json as std_json
from collections import Iterable
from datetime import date, datetime

from fauxfactory import gen_choice
from inflection import pluralize
from nailgun import client, config, signals
from nailgun.entity_fields import (
    IntegerField, OneToManyField, OneToOneField, ListField)
import threading
import time

from sys import version_info
if version_info.major == 2:  # pragma: no cover
    # pylint:disable=import-error
    from urlparse import urljoin
    import httplib as http_client
    import thread
else:  # pragma: no cover
    from urllib.parse import urljoin  # pylint:disable=F0401,E0611
    import _thread as thread  # pylint:disable=import-error
    import http.client as http_client  # pylint:disable=import-error

# pylint:disable=too-many-lines
# This module contains very extensive docstrings, so this module is easier to
# understand than its size suggests. That said, it could be useful to split
# each mixin in to a separate module. That would help to ensure that each
# mixin stays independent.


#: Default for ``poll_rate`` argument to
#: :func:`nailgun.entity_mixins._poll_task`.
TASK_POLL_RATE = 5
#: Default for ``timeout`` argument to
#: :func:`nailgun.entity_mixins._poll_task`.
TASK_TIMEOUT = 300

#: A :class:`nailgun.config.ServerConfig` object.
#:
#: Used by :class:`nailgun.entity_mixins.Entity`.
DEFAULT_SERVER_CONFIG = None

#: Used by :meth:`nailgun.entity_mixins.EntityCreateMixin.create_raw`.
#:
#: This is the default value for the ``create_missing`` argument to
#: :meth:`nailgun.entity_mixins.EntityCreateMixin.create_raw`. Keep in mind
#: that this variable also affects methods which call ``create_raw``, such as
#: :meth:`nailgun.entity_mixins.EntityCreateMixin.create_json`.
CREATE_MISSING = False


class TaskTimedOutError(Exception):
    """Indicates that a task did not finish before the timout limit."""


class TaskFailedError(Exception):
    """Indicates that a task finished with a result other than "success"."""


def _poll_task(task_id, server_config, poll_rate=None, timeout=None):
    """Implement :meth:`nailgun.entities.ForemanTask.poll`.

    See :meth:`nailgun.entities.ForemanTask.poll` for a full description of how
    this method acts. Other methods may also call this method, such as
    :meth:`nailgun.entity_mixins.EntityDeleteMixin.delete`.

    Certain mixins benefit from being able to poll the server after performing
    an operation. However, this module cannot use
    :meth:`nailgun.entities.ForemanTask.poll`, as that would be a circular
    import. Placing the implementation of
    :meth:`nailgun.entities.ForemanTask.poll` here allows both that method and
    the mixins in this module to use the same logic.

    """
    if poll_rate is None:
        poll_rate = TASK_POLL_RATE
    if timeout is None:
        timeout = TASK_TIMEOUT

    # Implement the timeout.
    def raise_task_timeout():  # pragma: no cover
        """Raise a KeyboardInterrupt exception in the main thread."""
        thread.interrupt_main()

    timer = threading.Timer(timeout, raise_task_timeout)

    # Poll until the task finishes. The timeout prevents an infinite loop.
    path = '{0}/foreman_tasks/api/tasks/{1}'.format(server_config.url, task_id)
    try:
        timer.start()
        while True:
            response = client.get(path, **server_config.get_client_kwargs())
            response.raise_for_status()
            task_info = response.json()
            if task_info['state'] in ('paused', 'stopped'):
                break
            time.sleep(poll_rate)
    except KeyboardInterrupt:  # pragma: no cover
        # raise_task_timeout will raise a KeyboardInterrupt when the timeout
        # expires. Catch the exception and raise TaskTimedOutError
        raise TaskTimedOutError(
            'Timed out polling task {0}. Task information: {1}'
            .format(task_id, task_info)
        )
    finally:
        timer.cancel()

    # Check for task success or failure.
    if task_info['result'] != 'success':
        raise TaskFailedError(
            'Task {0} did not succeed. Task information: {1}'
            .format(task_id, task_info)
        )
    return task_info


def _make_entity_from_id(entity_cls, entity_obj_or_id, server_config):
    """Given an entity object or an ID, return an entity object.

    If the value passed in is an object that is a subclass of :class:`Entity`,
    return that value. Otherwise, create an object of the type that ``field``
    references, give that object an ID of ``field_value``, and return that
    object.

    :param entity_cls: An :class:`Entity` subclass.
    :param entity_obj_or_id: Either a :class:`nailgun.entity_mixins.Entity`
        object or an entity ID.
    :returns: An ``entity_cls`` object.
    :rtype: nailgun.entity_mixins.Entity

    """
    if isinstance(entity_obj_or_id, entity_cls):
        return entity_obj_or_id
    return entity_cls(server_config, id=entity_obj_or_id)


def _make_entities_from_ids(entity_cls, entity_objs_and_ids, server_config):
    """Given an iterable of entities and/or IDs, return a list of entities.

    :param entity_cls: An :class:`Entity` subclass.
    :param entity_obj_or_id: An iterable of
        :class:`nailgun.entity_mixins.Entity` objects and/or entity IDs. All of
        the entities in this iterable should be of type ``entity_cls``.
    :returns: A list of ``entity_cls`` objects.

    """
    return [
        _make_entity_from_id(entity_cls, entity_or_id, server_config)
        for entity_or_id
        in entity_objs_and_ids
    ]


def _payload(fields, values):
    """Implement the ``*_payload`` methods.

    It's frequently useful to create a dict of values that can be encoded to
    JSON and sent to the server. Unfortunately, there are mismatches between
    the field names used by NailGun and the field names the server expects.
    This method provides a default translation that works in many cases. For
    example:

    >>> from nailgun.entities import Product
    >>> product = Product(name='foo', organization=1)
    >>> set(product.get_fields())
    {
        'description',
        'gpg_key',
        'id',
        'label',
        'name',
        'organization',
        'sync_plan',
    }
    >>> set(product.get_values())
    {'name', 'organization'}
    >>> product.create_payload()
    {'organization_id': 1, 'name': 'foo'}

    :param fields: A value like what is returned by
        :meth:`nailgun.entity_mixins.Entity.get_fields`.
    :param values: A value like what is returned by
        :meth:`nailgun.entity_mixins.Entity.get_values`.
    :returns: A dict mapping field names to field values.

    """
    for field_name, field in fields.items():
        if field_name in values:
            if isinstance(field, OneToOneField):
                values[field_name + '_id'] = (
                    getattr(values.pop(field_name), 'id', None)
                )
            elif isinstance(field, OneToManyField):
                values[field_name + '_ids'] = [
                    entity.id for entity in values.pop(field_name)
                ]
            elif isinstance(field, ListField):
                def parse(obj):
                    """parse obj payload if it is an Entity"""
                    if isinstance(obj, Entity):
                        return _payload(obj.get_fields(), obj.get_values())
                    return obj

                values[field_name] = [
                    parse(obj) for obj in values[field_name]]
    return values


def _get_server_config():
    """Search for a :class:`nailgun.config.ServerConfig`.

    :returns: :data:`nailgun.entity_mixins.DEFAULT_SERVER_CONFIG` if it is not
        ``None``, or whatever is returned by
        :meth:`nailgun.config.ServerConfig.get` otherwise.
    :rtype: nailgun.config.ServerConfig

    """
    if DEFAULT_SERVER_CONFIG is not None:
        return DEFAULT_SERVER_CONFIG
    return config.ServerConfig.get()


def _get_entity_id(field_name, attrs):
    """Find the ID for a one to one relationship.

    The server may return JSON data in the following forms for a
    :class:`nailgun.entity_fields.OneToOneField`::

        'user': None
        'user': {'name': 'Alice Hayes', 'login': 'ahayes', 'id': 1}
        'user_id': 1
        'user_id': None

    Search ``attrs`` for a one to one ``field_name`` and return its ID.

    :param field_name: A string. The name of a field.
    :param attrs: A dict. A JSON payload as returned from a server.
    :returns: Either an entity ID or None.

    """
    field_name_id = field_name + '_id'
    if field_name in attrs:
        if attrs[field_name] is None:
            return None
        else:
            return attrs[field_name]['id']
    elif field_name_id in attrs:
        return attrs[field_name_id]
    else:
        raise MissingValueError(
            'Cannot find a value for the "{0}" field. Searched for keys named '
            '{1}, but available keys are {2}.'
            .format(field_name, (field_name, field_name_id), attrs.keys())
        )


def _get_entity_ids(field_name, attrs):
    """Find the IDs for a one to many relationship.

    The server may return JSON data in the following forms for a
    :class:`nailgun.entity_fields.OneToManyField`::

        'user': [{'id': 1, …}, {'id': 42, …}]
        'users': [{'id': 1, …}, {'id': 42, …}]
        'user_ids': [1, 42]

    Search ``attrs`` for a one to many ``field_name`` and return its ID.

    :param field_name: A string. The name of a field.
    :param attrs: A dict. A JSON payload as returned from a server.
    :returns: An iterable of entity IDs.

    """
    field_name_ids = field_name + '_ids'
    plural_field_name = pluralize(field_name)
    if field_name_ids in attrs:
        return attrs[field_name_ids]
    elif field_name in attrs:
        return [entity['id'] for entity in attrs[field_name]]
    elif plural_field_name in attrs:
        return [entity['id'] for entity in attrs[plural_field_name]]
    else:
        raise MissingValueError(
            'Cannot find a value for the "{0}" field. Searched for keys named '
            '{1}, but available keys are {2}.'
            .format(
                field_name,
                (field_name_ids, field_name, plural_field_name),
                attrs.keys()
            )
        )


# -----------------------------------------------------------------------------
# Definition of parent Entity class and its dependencies.
# -----------------------------------------------------------------------------


class NoSuchPathError(Exception):
    """Indicates that the requested path cannot be constructed."""


class NoSuchFieldError(Exception):
    """Indicates that the referenced field does not exist."""


class BadValueError(Exception):
    """Indicates that an inappropriate value was assigned to an entity."""


class MissingValueError(Exception):
    """Indicates that no value can be found for a field."""


class Entity(object):
    """A representation of a logically related set of API paths.

    This class is rather useless as is, and it is intended to be subclassed.
    Subclasses can specify two useful types of information:

    * fields
    * metadata

    Fields and metadata are represented by the ``_fields`` and ``_meta``
    instance attributes, respectively. Here is an example of how to define and
    instantiate an entity:

    >>> class User(Entity):
    ...     def __init__(self, server_config=None, **kwargs):
    ...         self._fields = {
    ...             'name': StringField(),
    ...             'supervisor': OneToOneField('User'),
    ...             'subordinate': OneToManyField('User'),
    ...         }
    ...         self._meta = {'api_path': 'api/users'}
    ...         return super(User, self).__init__(server_config, **kwargs)
    ...
    >>> user = User(
    ...     name='Alice',
    ...     supervisor=User(id=1),
    ...     subordinate=[User(id=3), User(id=4)],
    ... )
    >>> user.name == 'Alice'
    True
    >>> user.supervisor.id = 1
    True

    The canonical procedure for initializing foreign key fields, shown above,
    is powerful but verbose. It is tiresome to write statements such as
    ``[User(id=3), User(id=4)]``. As a convenience, entity IDs may be given:

    >>> User(name='Alice', supervisor=1, subordinate=[3, 4])
    >>> user.name == 'Alice'
    True
    >>> user.supervisor.id = 1
    True

    An entity object is useless if you are unable to use it to communicate with
    a server. The solution is to provide a :class:`nailgun.config.ServerConfig`
    when instantiating a new entity.

    1. If the ``server_config`` argument is specified, then that is used.
    2. Otherwise, if :data:`nailgun.entity_mixins.DEFAULT_SERVER_CONFIG` is
       set, then that is used.
    3. Otherwise, call :meth:`nailgun.config.ServerConfig.get`.

    An entity's server configuration is stored as a private instance variaable
    and is used by mixin methods, such as
    :meth:`nailgun.entity_mixins.Entity.path`. For more information on server
    configuration objects, see :class:`nailgun.config.BaseServerConfig`.

    :raises nailgun.entity_mixins.NoSuchFieldError: If a value is assigned to a
        non-existent field.
    :raises nailgun.entity_mixins.BadValueError: If an inappropriate value is
        assigned to a field.

    """

    def __init__(self, server_config=None, **kwargs):
        if server_config is None:
            server_config = _get_server_config()
        self._server_config = server_config

        # Subclasses usually define a set of fields and metadata before calling
        # `super`, but that's not always the case.
        if not hasattr(self, '_fields'):
            self._fields = {}
        self._fields.setdefault('id', IntegerField())
        if not hasattr(self, '_meta'):
            self._meta = {}

        # Check that a valid set of field values has been passed in.
        if not set(kwargs.keys()).issubset(self._fields.keys()):
            raise NoSuchFieldError(
                'Valid fields are {0}, but received {1} instead.'
                .format(self._fields.keys(), kwargs.keys())
            )

        # Iterate through the values passed in and assign them as instance
        # variable to `self`. Make sure to transform entity IDs into entity
        # objects. (This feature is described in the docstring.)
        for field_name, field_value in kwargs.items():  # e.g. ('admin', True)
            field = self._fields[field_name]  # e.g. A BooleanField object
            if isinstance(field, OneToOneField):
                if field_value is None:
                    setattr(self, field_name, field_value)
                else:
                    setattr(self, field_name, _make_entity_from_id(
                        field.gen_value(),
                        field_value,
                        self._server_config
                    ))
            elif isinstance(field, OneToManyField):
                # `try:; …; except TypeError:; raise BadValueError(…)` better
                # follows the "ask forgiveness" principle. However, a TypeError
                # could be raised for any number of reasons. For example,
                # `field_value` could have a faulty __iter__ implementation.
                if not isinstance(field_value, Iterable):
                    raise BadValueError(
                        'An inappropriate value was assigned to the "{0}" '
                        'field. An iterable of entities and/or entity IDs '
                        'should be assigned, but the following was given: {1}'
                        .format(field_name, field_value)
                    )
                setattr(self, field_name, _make_entities_from_ids(
                    field.gen_value(),
                    field_value,
                    self._server_config
                ))
            else:
                setattr(self, field_name, field_value)

    def path(self, which=None):
        """Return the path to the current entity.

        Return the path to base entities of this entity's type if:

        * ``which`` is ``'base'``, or
        * ``which`` is ``None`` and instance attribute ``id`` is unset.

        Return the path to this exact entity if instance attribute ``id`` is
        set and:

        * ``which`` is ``'self'``, or
        * ``which`` is ``None``.

        Raise :class:`NoSuchPathError` otherwise.

        Child classes may choose to extend this method, especially if a child
        entity offers more than the two URLs supported by default. If extended,
        then the extending class should check for custom parameters before
        calling ``super``::

            def path(self, which):
                if which == 'custom':
                    return urljoin(…)
                super(ChildEntity, self).__init__(which)

        This will allow the extending method to accept a custom parameter
        without accidentally raising a :class:`NoSuchPathError`.

        :param which: A string. Optional. Valid arguments are 'self' and
            'base'.
        :return: A string. A fully qualified URL.
        :raises nailgun.entity_mixins.NoSuchPathError: If no path can be built.

        """
        # It is OK that member ``self._meta`` is not found. Subclasses are
        # required to set that attribute if they wish to use this method.
        #
        # Beware of leading and trailing slashes:
        #
        #     urljoin('example.com', 'foo') => 'foo'
        #     urljoin('example.com/', 'foo') => 'example.com/foo'
        #     urljoin('example.com', '/foo') => '/foo'
        #     urljoin('example.com/', '/foo') => '/foo'
        #
        base = urljoin(
            self._server_config.url + '/',
            self._meta['api_path']  # pylint:disable=no-member
        )
        if which == 'base' or (which is None and not hasattr(self, 'id')):
            return base
        elif (which == 'self' or which is None) and hasattr(self, 'id'):
            return urljoin(base + '/', str(self.id))  # pylint:disable=E1101
        raise NoSuchPathError

    def get_fields(self):
        """Return a copy of the fields on the current object.

        :return: A dict mapping field names to
            :class`nailgun.entity_fields.Field` objects.

        """
        return self._fields.copy()

    def get_values(self):
        """Return a copy of field values on the current object.

        This method is almost identical to ``vars(self).copy()``. However,
        only instance attributes that correspond to a field are included in
        the returned dict.

        :return: A dict mapping field names to user-provided values.
        """
        attrs = vars(self).copy()
        attrs.pop('_server_config')
        attrs.pop('_fields')
        attrs.pop('_meta')
        return attrs

    def __repr__(self):
        return '{0}.{1}({2}{3})'.format(
            self.__module__,
            type(self).__name__,
            repr(self._server_config),
            ''.join(
                ', {0}={1}'.format(key, repr(value))
                for key, value
                in self.get_values().items()
            )
        )

    def to_json(self):
        r"""Create a JSON encoded string with Entity properties. Ex:

        >>> from nailgun import entities, config
        >>> kwargs = {
        ...         'id': 1,
        ...         'name': 'Nailgun Org',
        ...     }
        >>> org = entities.Organization(config.ServerConfig('foo'), \*\*kwargs)
        >>> org.to_json()
        '{"id": 1, "name": "Nailgun Org"}'

        :return: str
        """
        return std_json.dumps(self.to_json_dict())

    def to_json_dict(self):
        """Create a dct with Entity properties for json encoding.
        It can be overridden by subclasses for each standard serialization
        doesn't work. By default it call _to_json_dict on OneToOne fields
        and build a list calling the same method on each object on OneToMany
        fields.

        :return: dct
        """
        fields, values = self.get_fields(), self.get_values()
        json_dct = {}
        for field_name, field in fields.items():
            if field_name in values:
                value = values[field_name]
                if value is None:
                    json_dct[field_name] = None
                    # This conditions is needed because some times you get
                    # None on an OneToOneField what lead to an error
                    # on bellow condition, e.g., calling value.to_json_dict()
                    # when value is None
                elif isinstance(field, OneToOneField):
                    json_dct[field_name] = value.to_json_dict()
                elif isinstance(field, OneToManyField):
                    json_dct[field_name] = [
                        entity.to_json_dict() for entity in value
                    ]
                else:
                    json_dct[field_name] = to_json_serializable(value)
        return json_dct

    def __eq__(self, other):
        """Compare two entities based on their properties. Even nested
        objects are considered for equality

        :param other: entity to compare self to
        :return: boolean indicating if entities are equal or not
        """
        if other is None:
            return False
        return self.to_json_dict() == other.to_json_dict()


class EntityDeleteMixin(object):
    """This mixin provides the ability to delete an entity.

    The methods provided by this class work together. The call tree looks like
    this::

        delete → delete_raw

    In short, here is what the methods do:

    :meth:`delete_raw`
        Make an HTTP DELETE request to the server.
    :meth:`delete`
        Check the server's response for errors and decode the response.

    """

    def delete_raw(self):
        """Delete the current entity.

        Make an HTTP DELETE call to ``self.path('base')``. Return the response.

        :return: A ``requests.response`` object.

        """
        return client.delete(
            self.path(which='self'),
            **self._server_config.get_client_kwargs()
        )

    @signals.emit(sender=signals.SENDER_CLASS, post_result_name='result')
    def delete(self, synchronous=True):
        """Delete the current entity.

        Call :meth:`delete_raw` and check for an HTTP 4XX or 5XX response.
        Return either the JSON-decoded response or information about a
        completed foreman task.

        :param synchronous: A boolean. What should happen if the server returns
            an HTTP 202 (accepted) status code? Wait for the task to complete
            if ``True``. Immediately return a response otherwise.
        :returns: A dict. Either the JSON-decoded response or information about
            a foreman task.
        :raises: ``requests.exceptions.HTTPError`` if the response has an HTTP
            4XX or 5XX status code.
        :raises: ``ValueError`` If an HTTP 202 response is received and the
            response JSON can not be decoded.
        :raises nailgun.entity_mixins.TaskTimedOutError: If an HTTP 202
            response is received, ``synchronous is True`` and the task times
            out.

        """

        response = self.delete_raw()
        response.raise_for_status()

        if (synchronous is True and
                response.status_code == http_client.ACCEPTED):
            return _poll_task(response.json()['id'], self._server_config)
        elif response.status_code == http_client.NO_CONTENT:
            # "The server successfully processed the request, but is not
            # returning any content. Usually used as a response to a successful
            # delete request."
            return
        return response.json()


class EntityReadMixin(object):
    """This mixin provides the ability to read an entity.

    The methods provided by this class work together. The call tree looks like
    this::

        read → read_json → read_raw

    In short, here is what the methods do:

    :meth:`read_raw`
        Make an HTTP GET request to the server.
    :meth:`read_json`
        Check the server's response for errors and decode the response.
    :meth:`read`
        Create a :class:`nailgun.entity_mixins.Entity` object representing the
        created entity and populate its fields with data returned from the
        server.

    See the individual methods for more detailed information.

    """

    def read_raw(self):
        """Get information about the current entity.

        Make an HTTP PUT call to ``self.path('self')``. Return the response.

        :return: A ``requests.response`` object.

        """
        return client.get(
            self.path('self'),
            **self._server_config.get_client_kwargs()
        )

    def read_json(self):
        """Get information about the current entity.

        Call :meth:`read_raw`. Check the response status code, decode JSON and
        return the decoded JSON as a dict.

        :return: A dict. The server's response, with all JSON decoded.
        :raises: ``requests.exceptions.HTTPError`` if the response has an HTTP
            4XX or 5XX status code.
        :raises: ``ValueError`` If the response JSON can not be decoded.

        """
        response = self.read_raw()
        response.raise_for_status()
        return response.json()

    def read(self, entity=None, attrs=None, ignore=None):
        """Get information about the current entity.

        1. Create a new entity of type ``type(self)``.
        2. Call :meth:`read_json` and capture the response.
        3. Populate the entity with the response.
        4. Return the entity.

        Step one is skipped if the ``entity`` argument is specified. Step two
        is skipped if the ``attrs`` argument is specified. Step three is
        modified by the ``ignore`` argument.

        All of an entity's one-to-one and one-to-many relationships are
        populated with objects of the correct type. For example, if
        ``SomeEntity.other_entity`` is a one-to-one relationship, this should
        return ``True``::

            isinstance(
                SomeEntity(id=N).read().other_entity,
                nailgun.entity_mixins.Entity
            )

        Additionally, both of these commands should succeed::

            SomeEntity(id=N).read().other_entity.id
            SomeEntity(id=N).read().other_entity.read().other_attr

        In the example above, ``other_entity.id`` is the **only** attribute
        with a meaningful value. Calling ``other_entity.read`` populates the
        remaining entity attributes.

        :param nailgun.entity_mixins.Entity entity: The object to be populated
            and returned. An object of type ``type(self)`` by default.
        :param attrs: A dict. Data used to populate the object's attributes.
            The response from
            :meth:`nailgun.entity_mixins.EntityReadMixin.read_json` by default.
        :param ignore: A set of attributes which should not be read from the
            server. This is mainly useful for attributes like a password which
            are not returned.
        :return: An instance of type ``type(self)``.
        :rtype: nailgun.entity_mixins.Entity

        """
        if entity is None:
            entity = type(self)(self._server_config)
        if attrs is None:
            attrs = self.read_json()
        if ignore is None:
            ignore = set()

        for field_name, field in entity.get_fields().items():
            if field_name in ignore:
                continue
            if isinstance(field, OneToOneField):
                entity_id = _get_entity_id(field_name, attrs)
                if entity_id is None:
                    referenced_entity = None
                else:
                    referenced_entity = field.entity(
                        self._server_config,
                        id=entity_id,
                    )
                setattr(entity, field_name, referenced_entity)
            elif isinstance(field, OneToManyField):
                referenced_entities = [
                    field.entity(self._server_config, id=entity_id)
                    for entity_id
                    in _get_entity_ids(field_name, attrs)
                ]
                setattr(entity, field_name, referenced_entities)
            else:
                setattr(entity, field_name, attrs[field_name])
        return entity


class EntityCreateMixin(object):
    """This mixin provides the ability to create an entity.

    The methods provided by this class work together. The call tree looks like
    this::

        create
        └── create_json
            └── create_raw
                ├── create_missing
                └── create_payload

    In short, here is what the methods do:

    :meth:`create_missing`
        Populate required fields with random values. Required fields that
        already have a value are not populated. This method is not called
        by default.
    :meth:`create_payload`
        Assemble a payload of data that can be encoded and sent to the server.
    :meth:`create_raw`
        Make an HTTP POST request to the server, including the payload.
    :meth:`create_json`
        Check the server's response for errors and decode the response.
    :meth:`create`
        Create a :class:`nailgun.entity_mixins.Entity` object representing the
        created entity and populate its fields with data returned from the
        server.

    See the individual methods for more detailed information.

    """

    def create_missing(self):
        """Automagically populate all required instance attributes.

        Iterate through the set of all required class
        :class:`nailgun.entity_fields.Field` defined on ``type(self)`` and
        create a corresponding instance attribute if none exists. Subclasses
        should override this method if there is some relationship between two
        required fields.

        :return: Nothing. This method relies on side-effects.

        """
        for field_name, field in self.get_fields().items():
            if field.required and not hasattr(self, field_name):
                # Most `gen_value` methods return a value such as an integer,
                # string or dictionary, but OneTo{One,Many}Field.gen_value
                # returns the referenced class.
                if hasattr(field, 'default'):
                    value = field.default
                elif hasattr(field, 'choices'):
                    value = gen_choice(field.choices)
                elif isinstance(field, OneToOneField):
                    value = field.gen_value()(self._server_config).create(True)
                elif isinstance(field, OneToManyField):
                    value = [
                        field.gen_value()(self._server_config).create(True)
                    ]
                else:
                    value = field.gen_value()
                setattr(self, field_name, value)

    def create_payload(self):
        """Create a payload of values that can be sent to the server.

        See :func:`_payload`.

        """
        return _payload(self.get_fields(), self.get_values())

    def create_raw(self, create_missing=None):
        """Create an entity.

        Possibly call :meth:`create_missing`. Then make an HTTP POST call to
        ``self.path('base')``. The request payload consists of whatever is
        returned by :meth:`create_payload`. Return the response.

        :param create_missing: Should :meth:`create_missing` be called? In
            other words, should values be generated for required, empty fields?
            Defaults to :data:`nailgun.entity_mixins.CREATE_MISSING`.
        :return: A ``requests.response`` object.

        """
        if create_missing is None:
            create_missing = CREATE_MISSING
        if create_missing:
            self.create_missing()
        return client.post(
            self.path('base'),
            self.create_payload(),
            **self._server_config.get_client_kwargs()
        )

    def create_json(self, create_missing=None):
        """Create an entity.

        Call :meth:`create_raw`. Check the response status code, decode JSON
        and return the decoded JSON as a dict.

        :return: A dict. The server's response, with all JSON decoded.
        :raises: ``requests.exceptions.HTTPError`` if the response has an HTTP
            4XX or 5XX status code.
        :raises: ``ValueError`` If the response JSON can not be decoded.

        """
        response = self.create_raw(create_missing)
        response.raise_for_status()
        return response.json()

    @signals.emit(sender=signals.SENDER_CLASS, post_result_name='entity')
    def create(self, create_missing=None):
        """Create an entity.

        Call :meth:`create_json`, use the response to populate a new object
        of type ``type(self)`` and return that object.

        This method requires that a method named "read" be available on the
        current object. A method named "read" will be available if
        :class:`EntityReadMixin` is present in the inheritance tree, and using
        the method provided by that mixin is the recommended technique for
        making a "read" method available.

        This method makes use of :meth:`EntityReadMixin.read` for two reasons.
        First, calling that method is simply convenient. Second, the server
        frequently returns weirdly structured, inconsistently named or
        straight-up broken responses, and quite a bit of effort has gone in to
        decoding server responses so :meth:`EntityReadMixin.read` can function
        correctly. Calling ``read`` allows this method to re-use the decoding
        work that has been done for that method.

        :return: An instance of type ``type(self)``.
        :rtype: nailgun.entity_mixins.Entity
        :raises: ``AttributeError`` if a method named "read" is not available
            on the current object.

        """
        return self.read(attrs=self.create_json(create_missing))


class EntityUpdateMixin(object):
    """This mixin provides the ability to update an entity.

    The methods provided by this class work together. The call tree looks
    like this::

        update → update_json → update_raw → update_payload

    In short, here is what the methods do:

    :meth:`update_payload`
        Assemble a payload of data that can be encoded and sent to the
        server.
    :meth:`update_raw`
        Make an HTTP PUT request to the server, including the payload.
    :meth:`update_json`
        Check the server's response for errors and decode the response.
    :meth:`update`
        Create a :class:`nailgun.entity_mixins.Entity` object representing
        the created entity and populate its fields.

    See the individual methods for more detailed information.

    """

    def update_payload(self, fields=None):
        """Create a payload of values that can be sent to the server.

        By default, this method behaves just like :func:`_payload`. However,
        one can also specify a certain set of fields that should be returned.
        For more information, see :meth:`update`.

        """
        values = self.get_values()
        if fields is not None:
            values = {field: values[field] for field in fields}
        return _payload(self.get_fields(), values)

    def update_raw(self, fields=None):
        """Update the current entity.

        Make an HTTP PUT call to ``self.path('base')``. The request payload
        consists of whatever is returned by :meth:`update_payload`. Return the
        response.

        :param fields: See :meth:`update`.
        :return: A ``requests.response`` object.

        """
        return client.put(
            self.path('self'),
            self.update_payload(fields),
            **self._server_config.get_client_kwargs()
        )

    def update_json(self, fields=None):
        """Update the current entity.

        Call :meth:`update_raw`. Check the response status code, decode JSON
        and return the decoded JSON as a dict.

        :param fields: See :meth:`update`.
        :return: A dict consisting of the decoded JSON in the server's
            response.
        :raises: ``requests.exceptions.HTTPError`` if the response has an HTTP
            4XX or 5XX status code.
        :raises: ``ValueError`` If the response JSON can not be decoded.

        """
        response = self.update_raw(fields)
        response.raise_for_status()
        return response.json()

    @signals.emit(sender=signals.SENDER_CLASS, post_result_name='entity')
    def update(self, fields=None):
        """Update the current entity.

        Call :meth:`update_json`, use the response to populate a new object
        of type ``type(self)`` and return that object.

        This method requires that
        :meth:`nailgun.entity_mixins.EntityReadMixin.read` or some other
        identical method be available on the current object. A more thorough
        explanation is available at
        :meth:`nailgun.entity_mixins.EntityCreateMixin.create`.

        :param fields: An iterable of field names. Only the fields named in
            this iterable will be updated. No fields are updated if an empty
            iterable is passed in. All fields are updated if ``None`` is passed
            in.
        :raises: ``KeyError`` if asked to update a field but no value is
            available for that field on the current entity.

        """
        return self.read(attrs=self.update_json(fields))


class EntitySearchMixin(object):
    """This mixin provides the ability to search for entities.

    The methods provided by this class work together. The call tree looks like
    this::

        search
        ├── search_json
        │   └── search_raw
        │       └── search_payload
        ├── search_normalize
        └── search_filter

    In short, here is what the methods do:

    :meth:`search_payload`
        Assemble a search query that can be encoded and sent to the server.
    :meth:`search_raw`
        Make an HTTP GET request to the server, including the payload.
    :meth:`search_json`
        Check the server's response for errors and decode the response.
    :meth:`search_normalize`
        Normalize search results so they can be used to create new entities.
    :meth:`search`
        Create one or more :class:`nailgun.entity_mixins.Entity` objects
        representing the found entities and populate their fields.
    :meth:`search_filter`
        Read all ``entities`` and locally filter them.

    See the individual methods for more detailed information.

    """

    def search_payload(self, fields=None, query=None):
        """Create a search query.

        Do the following:

        1. Generate a search query. By default, all values returned by
           :meth:`nailgun.entity_mixins.Entity.get_values` are used. If
           ``fields`` is specified, only the named values are used.
        2. Merge ``query`` in to the generated search query.
        3. Return the result.

        The rules for generating a search query can be illustrated by example.
        Let's say that we have an entity with an
        :class:`nailgun.entity_fields.IntegerField`, a
        :class:`nailgun.entity_fields.OneToOneField` and a
        :class:`nailgun.entity_fields.OneToManyField`::

            >>> some_entity = SomeEntity(id=1, one=2, many=[3, 4])
            >>> fields = some_entity.get_fields()
            >>> isinstance(fields['id'], IntegerField)
            True
            >>> isinstance(fields['one'], OneToOneField)
            True
            >>> isinstance(fields['many'], OneToManyField)
            True

        This method appends "_id" and "_ids" on to the names of each
        ``OneToOneField`` and ``OneToManyField``, respectively::

            >>> some_entity.search_payload()
            {'id': 1, 'one_id': 2, 'many_ids': [3, 4]}

        By default, all fields are used. But you can specify a set of field
        names to use::

            >>> some_entity.search_payload({'id'})
            {'id': 1}
            >>> some_entity.search_payload({'one'})
            {'one_id': 2}
            >>> some_entity.search_payload({'id', 'one'})
            {'id': 1, 'one_id': 2}

        If a ``query`` is specified, it is merged in to the generated query::

            >>> some_entity.search_payload(query={'id': 5})
            {'id': 5, 'one_id': 2, 'many_ids': [3, 4]}
            >>> some_entity.search_payload(query={'per_page': 1000})
            {'id': 1, 'one_id': 2, 'many_ids': [3, 4], 'per_page': 1000}

        .. WARNING:: This method currently generates an extremely naive search
            query that will be wrong in many cases. In addition, Satellite
            currently accepts invalid search queries without complaint. Make
            sure to check the API documentation for your version of Satellite
            against what this method produces.

        :param fields: See :meth:`search`.
        :param query: See :meth:`search`.
        :returns: A dict that can be encoded as JSON and used in a search.

        """
        if fields is None:
            fields = set(self.get_values().keys())
        if query is None:
            query = {}

        payload = {}
        fields_dict = self.get_fields()
        for field in fields:
            value = getattr(self, field)
            if isinstance(fields_dict[field], OneToOneField):
                payload[field + '_id'] = value.id
            elif isinstance(fields_dict[field], OneToManyField):
                payload[field + '_ids'] = [entity.id for entity in value]
            else:
                payload[field] = value
        payload.update(query)
        return payload

    def search_raw(self, fields=None, query=None):
        """Search for entities.

        Make an HTTP GET call to ``self.path('base')``. Return the response.

        .. WARNING:: Subclasses that override this method should not alter the
            ``fields`` or ``query`` arguments. (However, subclasses that
            override this method may still alter the server's response.) See
            :meth:`search_normalize` for details.

        :param fields: See :meth:`search`.
        :param query: See :meth:`search`.
        :return: A ``requests.response`` object.

        """
        return client.get(
            self.path('base'),
            data=self.search_payload(fields, query),
            **self._server_config.get_client_kwargs()
        )

    def search_json(self, fields=None, query=None):
        """Search for entities.

        Call :meth:`search_raw`. Check the response status code, decode JSON
        and return the decoded JSON as a dict.

        .. WARNING:: Subclasses that override this method should not alter the
            ``fields`` or ``query`` arguments. (However, subclasses that
            override this method may still alter the server's response.) See
            :meth:`search_normalize` for details.

        :param fields: See :meth:`search`.
        :param query: See :meth:`search`.
        :return: A dict. The server's response, with all JSON decoded.
        :raises: ``requests.exceptions.HTTPError`` if the response has an HTTP
            4XX or 5XX status code.
        :raises: ``ValueError`` If the response JSON can not be decoded.

        """
        response = self.search_raw(fields, query)
        response.raise_for_status()
        return response.json()

    def search_normalize(self, results):
        """Normalize search results so they can be used to create new entities.

        See :meth:`search` for an example of how to use this method. Here's a
        simplified example::

            results = self.search_json()
            results = self.search_normalize(results)
            entity = SomeEntity(some_cfg, **results[0])

        At this time, it is possible to parse all search results without
        knowing what search query was sent to the server. However, it is
        possible that certain responses can only be parsed if the search query
        is known. If that is the case, this method will be given a new
        ``payload`` argument, where ``payload`` is the query sent to the
        server.

        As a precaution, the following is higly recommended:

        * :meth:`search` may alter ``fields`` and ``query`` at will.
        * :meth:`search_payload` may alter ``fields`` and ``query`` in an
          idempotent manner.
        * No other method should alter ``fields`` or ``query``.

        :param results: A list of dicts, where each dict is a set of attributes
            for one entity. The contents of these dicts are as is returned from
            the server.
        :returns: A list of dicts, where each dict is a set of attributes for
            one entity. The contents of these dicts have been normalized and
            can be used to instantiate entities.

        """
        fields = self.get_fields()
        normalized = []
        for result in results:
            # For each field that we know about, copy the corresponding field
            # from the server's search result. If any extra attributes are
            # copied over, Entity.__init__ will raise a NoSuchFieldError.
            # Examples of problematic results from server:
            #
            # * organization_id (denormalized OneToOne. see above)
            # * organizations, organization_ids (denormalized OneToMany. above)
            # * updated_at, created_at (these may be handled in the future)
            # * sp_subnet (Host.sp_subnet is an undocumented field)
            #
            attrs = {}
            for field_name, field in fields.items():
                if isinstance(field, OneToOneField):
                    try:
                        attrs[field_name] = _get_entity_id(field_name, result)
                    except MissingValueError:
                        pass
                elif isinstance(field, OneToManyField):
                    try:
                        attrs[field_name] = _get_entity_ids(field_name, result)
                    except MissingValueError:
                        pass
                else:
                    try:
                        attrs[field_name] = result[field_name]
                    except KeyError:
                        pass
            normalized.append(attrs)
        return normalized

    @signals.emit(sender=signals.SENDER_CLASS, post_result_name='entities')
    def search(self, fields=None, query=None, filters=None):
        """Search for entities.

        At its simplest, this method searches for all entities of a given kind.
        For example, to ask for all
        :class:`nailgun.entities.LifecycleEnvironment` entities::

            LifecycleEnvironment().search()

        Values on an entity are used to generate a search query, and the
        ``fields`` argument can be used to specify which fields should be used
        when generating a search query::

            lc_env = LifecycleEnvironment(name='foo', organization=1)
            results = lc_env.search()  # Search by name and organization.
            results = lc_env.search({'name', 'organization'})  # Same.
            results = lc_env.search({'name'})  # Search by name.
            results = lc_env.search({'organization'})  # Search by organization
            results = lc_env.search(set())  # Search for all lifecycle envs.
            results = lc_env.search({'library'})  # Error!

        In some cases, the simple search queries that can be generated by
        NailGun are not sufficient. In this case, you can pass in a raw search
        query instead. For example, to search for all lifecycle environments
        with a name of 'foo'::

            LifecycleEnvironment().search(query={'search': 'name="foo"'})

        The example above is rather pointless: it is easier and more concise to
        use a generated query. But — and this is a **very** important "but" —
        the manual search query is melded in to the generated query. This can
        be used to great effect::

            LifecycleEnvironment(name='foo').search(query={'per_page': 50})

        For examples of what the final search queries look like, see
        :meth:`search_payload`. (That method also accepts the ``fields`` and
        ``query`` arguments.)

        In some cases, the server's search facilities may be insufficient, or
        it may be inordinately difficult to craft a search query. In this case,
        you can filter search results locally. For example, to ask the server
        for a list of all lifecycle environments and then locally search
        through the results for the lifecycle environment named "foo"::

            LifecycleEnvironment().search(filters={'name': 'foo'})

        Be warned that filtering locally can be **very** slow. NailGun must
        ``read()`` every single entity returned by the server before filtering
        results. This is because the values used in the filtering process may
        not have been returned by the server in the initial response to the
        search.

        The fact that all entities are read when ``filters`` is specified can
        be used to great effect. For example, this search returns a fully
        populated list of every single lifecycle environment::

            LifecycleEnvironment().search(filters={})

        :param fields: A set naming which fields should be used when generating
            a search query. If ``None``, all values on the entity are used. If
            an empty set, no values are used.
        :param query: A dict containing a raw search query. This is melded in
            to the generated search query like so:  ``{generated:
            query}.update({manual: query})``.
        :param filters: A dict. Used to filter search results locally.
        :return: A list of entities, all of type ``type(self)``.

        """
        # Goals:
        #
        # * Be tolerant of missing values. It's reasonable for the server to
        #   return an incomplete set of attributes for each search result.
        # * Use as many returned values as possible. There's no point in
        #   letting returned data go to waste. This implies that we must…
        # * …parse irregular server responses. This includes pluralized field
        #   names, misnamed attributes (e.g. BZ 1233245) and weirdly named
        #   fields (e.g. Media.path_).
        #
        results = self.search_json(fields, query)['results']
        results = self.search_normalize(results)
        entities = [
            type(self)(self._server_config, **result)
            for result in results
        ]
        if filters is not None:
            entities = self.search_filter(entities, filters)
        return entities

    @staticmethod
    def search_filter(entities, filters):
        """Read all ``entities`` and locally filter them.

        This method can be used like so::

            entities = EntitySearchMixin(entities, {'name': 'foo'})

        In this example, only entities where ``entity.name == 'foo'`` holds
        true are returned. An arbitrary number of field names and values may be
        provided as filters.

        .. NOTE:: This method calls :meth:`EntityReadMixin.read`. As a result,
            this method only works when called on a class that also inherits
            from :class:`EntityReadMixin`.

        :param entities: A list of :class:`Entity` objects. All list items
            should be of the same type.
        :param filters: A dict in the form ``{field_name: field_value, …}``.
        :raises nailgun.entity_mixins.NoSuchFieldError: If any of the fields
            named in ``filters`` do not exist on the entities being filtered.
        :raises: ``NotImplementedError`` If any of the fields named in
            ``filters`` are a :class:`nailgun.entity_fields.OneToOneField` or
            :class:`nailgun.entity_fields.OneToManyField`.

        """
        # Check to make sure all arguments are sane.
        if len(entities) == 0:
            return entities
        fields = entities[0].get_fields()  # assume all entities are identical
        if not set(filters).issubset(fields):
            raise NoSuchFieldError(
                'Valid filters are {0}, but received {1} instead.'
                .format(fields.keys(), filters.keys())
            )
        for field_name in filters:
            if isinstance(fields[field_name], (OneToOneField, OneToManyField)):
                raise NotImplementedError(
                    'Search results cannot (yet?) be locally filtered by '
                    '`OneToOneField`s and `OneToManyField`s. {0} is a {1}.'
                    .format(field_name, type(fields[field_name]).__name__)
                )

        # The arguments are sane. Filter away!
        filtered = [entity.read() for entity in entities]  # don't alter inputs
        for field_name, field_value in filters.items():
            filtered = [
                entity for entity in filtered
                if getattr(entity, field_name) == field_value
            ]
        return filtered


def to_json_serializable(obj):
    """ Transforms obj into a json serializable object.

    :param obj: entity or any json serializable object

    :return: serializable object

    """
    if isinstance(obj, Entity):
        return obj.to_json_dict()

    if isinstance(obj, dict):
        return {k: to_json_serializable(v) for k, v in obj.items()}
    elif isinstance(obj, (list, tuple)):
        return [to_json_serializable(v) for v in obj]
    elif isinstance(obj, datetime):
        return obj.strftime('%Y-%m-%d %H:%M:%S')
    elif isinstance(obj, date):
        return obj.strftime('%Y-%m-%d')

    return obj



# -*- coding: utf-8 -*-
"""Tests for :mod:`nailgun.entity_mixins`."""
# Python 3.3 and later includes module `ipaddress` in the standard library. If
# NailGun ever moves past Python 2.x, that module should be used instead of
# `socket`.
from fauxfactory import gen_integer
from nailgun import client, config, entity_mixins
from nailgun.entity_fields import (
    IntegerField,
    ListField,
    OneToManyField,
    OneToOneField,
    StringField,
)
from requests.exceptions import HTTPError
import mock

from sys import version_info
if version_info.major == 2:
    import httplib as http_client  # pylint:disable=import-error
else:
    import http.client as http_client  # pylint:disable=import-error
if version_info < (3, 4):
    from unittest2 import TestCase  # pylint:disable=import-error
else:
    from unittest import TestCase

# pylint:disable=too-many-lines
# The size of this module is a direct reflection of the size of module
# `nailgun.entity_mixins`. It would be good to split that module up, then split
# this module up similarly.

# This module is divided in to the following sections:
#
# 1. Entity defintions.
# 2. Tests for private methods.
# 3. Tests for public methods.
#
# 1. Entity definitions. ------------------------------------------------- {{{1


class SampleEntity(entity_mixins.Entity):
    """Sample entity to be used in the tests"""

    def __init__(self, server_config=None, **kwargs):
        self._fields = {'name': StringField(), 'number': IntegerField()}
        self._meta = {'api_path': 'foo'}
        super(SampleEntity, self).__init__(server_config, **kwargs)


class SampleEntityTwo(entity_mixins.Entity):
    """An entity with foreign key fields.

    This class has a :class:`nailgun.entity_fields.OneToManyField` called
    "one_to_many" pointing to :class:`tests.test_entity_mixins.SampleEntity`.

    """

    def __init__(self, server_config=None, **kwargs):
        self._fields = {'one_to_many': OneToManyField(SampleEntity)}
        super(SampleEntityTwo, self).__init__(server_config, **kwargs)


class SampleEntityThree(entity_mixins.Entity):
    """An entity with foreign key fields as One to One and ListField.

    This class has a :class:`nailgun.entity_fields.OneToOneField` called
    "one_to_one" pointing to :class:`tests.test_entity_mixins.SampleEntityTwo`.

    This class has a :class:`nailgun.entity_fields.ListField` called "list"
    containing instances of :class:`tests.test_entity_mixins.SampleEntity`.

    """

    def __init__(self, server_config=None, **kwargs):
        self._fields = {
            'one_to_one': OneToOneField(SampleEntityTwo),
            'list': ListField()
        }
        super(SampleEntityThree, self).__init__(server_config, **kwargs)


class EntityWithCreate(entity_mixins.Entity, entity_mixins.EntityCreateMixin):
    """Inherits from :class:`nailgun.entity_mixins.EntityCreateMixin`."""

    def __init__(self, server_config=None, **kwargs):
        self._meta = {'api_path': ''}
        super(EntityWithCreate, self).__init__(server_config, **kwargs)


class EntityWithRead(entity_mixins.Entity, entity_mixins.EntityReadMixin):
    """Inherits from :class:`nailgun.entity_mixins.EntityReadMixin`."""

    def __init__(self, server_config=None, **kwargs):
        self._meta = {'api_path': ''}
        super(EntityWithRead, self).__init__(server_config, **kwargs)


class EntityWithUpdate(entity_mixins.Entity, entity_mixins.EntityUpdateMixin):
    """Inherits from :class:`nailgun.entity_mixins.EntityUpdateMixin`."""

    def __init__(self, server_config=None, **kwargs):
        self._meta = {'api_path': ''}
        super(EntityWithUpdate, self).__init__(server_config, **kwargs)


class EntityWithDelete(entity_mixins.Entity, entity_mixins.EntityDeleteMixin):
    """Inherits from :class:`nailgun.entity_mixins.EntityDeleteMixin`."""

    def __init__(self, server_config=None, **kwargs):
        self._meta = {'api_path': ''}
        super(EntityWithDelete, self).__init__(server_config, **kwargs)


class EntityWithSearch(entity_mixins.Entity, entity_mixins.EntitySearchMixin):
    """Inherits from :class:`nailgun.entity_mixins.EntitySearchMixin`."""

    def __init__(self, server_config=None, **kwargs):
        self._meta = {'api_path': ''}
        super(EntityWithSearch, self).__init__(server_config, **kwargs)


class EntityWithSearch2(EntityWithSearch):
    """An entity with integer, one to one and one to many fields."""

    def __init__(self, server_config=None, **kwargs):
        self._fields = {
            'one': OneToOneField(SampleEntity),
            'many': OneToManyField(SampleEntity),
        }
        super(EntityWithSearch2, self).__init__(server_config, **kwargs)


# 2. Tests for private methods. ------------------------------------------ {{{1


class MakeEntityFromIdTestCase(TestCase):
    """Tests for :func:`nailgun.entity_mixins._make_entity_from_id`."""
    # pylint:disable=protected-access

    def setUp(self):
        """Set ``self.cfg``."""
        self.cfg = config.ServerConfig('example.com')

    def test_pass_in_entity_obj(self):
        """Let the ``entity_obj_or_id`` argument be an entity object."""
        self.assertIsInstance(
            entity_mixins._make_entity_from_id(
                SampleEntity,
                SampleEntity(self.cfg),
                self.cfg
            ),
            SampleEntity
        )

    def test_pass_in_entity_id(self):
        """Let the ``entity_obj_or_id`` argument be an integer."""
        entity_id = gen_integer(min_value=1)
        entity_obj = entity_mixins._make_entity_from_id(
            SampleEntity,
            entity_id,
            self.cfg
        )
        self.assertIsInstance(entity_obj, SampleEntity)
        self.assertEqual(entity_obj.id, entity_id)  # pylint:disable=no-member


class MakeEntitiesFromIdsTestCase(TestCase):
    """Tests for :func:`nailgun.entity_mixins._make_entities_from_ids`."""
    # pylint:disable=protected-access

    def setUp(self):
        """Set ``self.cfg``."""
        self.cfg = config.ServerConfig('example.com')

    def test_pass_in_emtpy_iterable(self):
        """Let the ``entity_objs_and_ids`` argument be an empty iterable."""
        for iterable in ([], tuple()):
            self.assertEqual(
                entity_mixins._make_entities_from_ids(
                    SampleEntity,
                    iterable,
                    self.cfg
                ),
                [],
            )

    def test_pass_in_entity_obj(self):
        """Let the ``entity_objs_and_ids`` arg be an iterable of entities."""
        for num_entities in range(4):
            input_entities = [
                SampleEntity(self.cfg) for _ in range(num_entities)
            ]
            output_entities = entity_mixins._make_entities_from_ids(
                SampleEntity,
                input_entities,
                self.cfg
            )
            self.assertEqual(num_entities, len(output_entities))
            for output_entity in output_entities:
                self.assertIsInstance(output_entity, SampleEntity)

    def test_pass_in_entity_ids(self):
        """Let the ``entity_objs_and_ids`` arg be an iterable of integers."""
        for num_entities in range(4):
            entity_ids = [
                gen_integer(min_value=1) for _ in range(num_entities)
            ]
            entities = entity_mixins._make_entities_from_ids(
                SampleEntity,
                entity_ids,
                self.cfg
            )
            self.assertEqual(len(entities), len(entity_ids))
            for i, entity_id in enumerate(entity_ids):
                self.assertIsInstance(entities[i], SampleEntity)
                self.assertEqual(entities[i].id, entity_id)

    def test_pass_in_both(self):
        """Let ``entity_objs_and_ids`` be an iterable of integers and IDs."""
        entities = entity_mixins._make_entities_from_ids(
            SampleEntity,
            [SampleEntity(self.cfg), 5],
            self.cfg
        )
        self.assertEqual(len(entities), 2)
        for entity in entities:
            self.assertIsInstance(entity, SampleEntity)


class PollTaskTestCase(TestCase):
    """Tests for :func:`nailgun.entity_mixins._poll_task`."""

    def setUp(self):
        """Create a bogus server configuration object."""
        self.cfg = config.ServerConfig('bogus url')

    def test__poll_task_failure(self):
        """What happens when a foreman task completes but does not succeed?

        Assert that a :class:`nailgun.entity_mixins.TaskFailedError` exception
        is raised.

        """
        for state in ('paused', 'stopped'):
            with self.subTest(state):
                with mock.patch.object(client, 'get') as get:
                    get.return_value.json.return_value = {
                        'state': state,
                        'result': 'not success'
                    }
                    with self.assertRaises(entity_mixins.TaskFailedError):
                        # pylint:disable=protected-access
                        entity_mixins._poll_task(gen_integer(), self.cfg)

    def test__poll_task_success(self):
        """What happens when a foreman task completes and does succeed?

        Assert that the server's response is returned.

        """
        for state in ('paused', 'stopped'):
            with self.subTest(state):
                with mock.patch.object(client, 'get') as get:
                    get.return_value.json.return_value = {
                        'state': state,
                        'result': 'success'
                    }
                    self.assertEqual(
                        get.return_value.json.return_value,
                        # pylint:disable=protected-access
                        entity_mixins._poll_task(gen_integer(), self.cfg),
                    )


# 3. Tests for public methods. ------------------------------------------- {{{1


class EntityTestCase(TestCase):
    """Tests for :class:`nailgun.entity_mixins.Entity`."""

    def setUp(self):
        """Set ``self.cfg``."""
        self.cfg = config.ServerConfig('http://example.com')

    def test_init_v1(self):
        """Provide no value for the ``server_config`` argument."""
        with mock.patch.object(config.ServerConfig, 'get') as sc_get:
            self.assertEqual(
                SampleEntity()._server_config,  # pylint:disable=W0212
                sc_get.return_value,
            )
        self.assertEqual(sc_get.call_count, 1)

    def test_init_v2(self):
        """Provide a server config object via ``DEFAULT_SERVER_CONFIG``."""
        backup = entity_mixins.DEFAULT_SERVER_CONFIG
        try:
            entity_mixins.DEFAULT_SERVER_CONFIG = config.ServerConfig('url')
            self.assertEqual(
                SampleEntity()._server_config,  # pylint:disable=W0212
                entity_mixins.DEFAULT_SERVER_CONFIG,
            )
        finally:
            entity_mixins.DEFAULT_SERVER_CONFIG = backup

    def test_entity_get_fields(self):
        """Test :meth:`nailgun.entity_mixins.Entity.get_fields`."""
        fields = SampleEntity(self.cfg).get_fields()
        self.assertEqual(len(fields), 3)
        self.assertEqual(set(fields.keys()), {'id', 'name', 'number'})
        self.assertIsInstance(fields['name'], StringField)
        self.assertIsInstance(fields['number'], IntegerField)

    def test_entity_get_values(self):
        """Test :meth:`nailgun.entity_mixins.Entity.get_values`."""
        for values in (
                {},
                {'id': gen_integer()},
                {'name': gen_integer()},
                {'number': gen_integer()},
                {'name': gen_integer(), 'number': gen_integer()},
                {
                    'id': gen_integer(),
                    'name': gen_integer(),
                    'number': gen_integer(),
                },
        ):
            self.assertEqual(
                SampleEntity(self.cfg, **values).get_values(),
                values,
            )

    def test_path(self):
        """Test :meth:`nailgun.entity_mixins.Entity.path`."""
        # e.g. 'https://sat.example.com/katello/api/v2'
        path = '{0}/{1}'.format(
            self.cfg.url,
            # pylint:disable=protected-access
            SampleEntity(self.cfg)._meta['api_path']
        )

        # Call `path()` on an entity with no ID.
        self.assertEqual(SampleEntity(self.cfg).path(), path)
        self.assertEqual(SampleEntity(self.cfg).path('base'), path)
        with self.assertRaises(entity_mixins.NoSuchPathError):
            SampleEntity(self.cfg).path('self')

        # Call `path()` on an entity with an ID.
        self.assertEqual(SampleEntity(self.cfg, id=5).path(), path + '/5')
        self.assertEqual(SampleEntity(self.cfg, id=5).path('base'), path)
        self.assertEqual(
            SampleEntity(self.cfg, id=5).path('self'), path + '/5')

    def test_no_such_field_error(self):
        """Try to raise a :class:`nailgun.entity_mixins.NoSuchFieldError`."""
        SampleEntity(self.cfg, name='Alice')
        with self.assertRaises(entity_mixins.NoSuchFieldError):
            SampleEntity(self.cfg, namee='Alice')

    def test_bad_value_error(self):
        """Try to raise a :class:`nailgun.entity_mixins.BadValueError`."""
        SampleEntityTwo(self.cfg, one_to_many=[1])
        with self.assertRaises(entity_mixins.BadValueError):
            SampleEntityTwo(self.cfg, one_to_many=1)

    def test_eq_none(self):
        """Test method ``nailgun.entity_mixins.Entity.__eq__`` against None

        Assert that ``__eq__`` returns False when compared to None.

        """
        alice = SampleEntity(self.cfg, id=1, name='Alice')
        self.assertFalse(alice.__eq__(None))

    def test_eq(self):
        """Test method ``nailgun.entity_mixins.Entity.__eq__``.

        Assert that ``__eq__`` works comparing all attributes, even from
        nested structures.

        """
        # Testing simple properties
        alice = SampleEntity(self.cfg, id=1, name='Alice')
        alice_clone = SampleEntity(self.cfg, id=1, name='Alice')
        self.assertEqual(alice, alice_clone)

        alice_id_2 = SampleEntity(self.cfg, id=2, name='Alice')
        self.assertNotEqual(alice, alice_id_2)

        # Testing OneToMany nested objects

        john = SampleEntityTwo(self.cfg, one_to_many=[alice, alice_id_2])
        john_clone = SampleEntityTwo(self.cfg, one_to_many=[alice, alice_id_2])
        self.assertEqual(john, john_clone)

        john_different_order = SampleEntityTwo(self.cfg, one_to_many=[
            alice_id_2, alice,
        ])
        self.assertNotEqual(john, john_different_order)

        john_missing_alice = SampleEntityTwo(self.cfg, one_to_many=[alice])
        self.assertNotEqual(john, john_missing_alice)

        john_without_alice = SampleEntityTwo(self.cfg)
        self.assertNotEqual(john, john_without_alice)

        # Testing OneToOne nested objects

        mary = SampleEntityThree(self.cfg, one_to_one=john)
        mary_clone = SampleEntityThree(
            self.cfg, one_to_one=john_clone)
        self.assertEqual(mary, mary_clone)

        mary_different = SampleEntityThree(
            self.cfg, one_to_one=john_different_order)
        self.assertNotEqual(mary, mary_different)

        mary_none_john = SampleEntityThree(self.cfg, one_to_one=None)
        mary_none_john.to_json_dict()
        self.assertNotEqual(mary, mary_none_john)

        # Testing List nested objects
        # noqa pylint:disable=attribute-defined-outside-init
        mary.list = [alice]
        self.assertNotEqual(mary, mary_clone)
        # noqa pylint:disable=attribute-defined-outside-init
        mary_clone.list = [alice_clone]
        self.assertEqual(mary, mary_clone)

    def test_repr_v1(self):
        """Test method ``nailgun.entity_mixins.Entity.__repr__``.

        Assert that ``__repr__`` works correctly when no arguments are passed
        to an entity.

        """
        entity = SampleEntityTwo(self.cfg)
        target = 'tests.test_entity_mixins.SampleEntityTwo({0})'.format(
            repr(self.cfg)
        )
        self.assertEqual(repr(entity), target)
        import nailgun  # noqa pylint:disable=unused-variable
        import tests  # noqa pylint:disable=unused-variable
        # pylint:disable=eval-used
        self.assertEqual(repr(eval(repr(entity))), target)

    def test_repr_v2(self):
        """Test method ``nailgun.entity_mixins.Entity.__repr__``.

        Assert that ``__repr__`` works correctly when an ID is passed to an
        entity.

        """
        entity = SampleEntityTwo(self.cfg, id=gen_integer())
        target = (
            'tests.test_entity_mixins.SampleEntityTwo({0}, id={1})'
            .format(repr(self.cfg), entity.id)  # pylint:disable=no-member
        )
        self.assertEqual(repr(entity), target)
        import nailgun  # noqa pylint:disable=unused-variable
        import tests  # noqa pylint:disable=unused-variable
        # pylint:disable=eval-used
        self.assertEqual(repr(eval(repr(entity))), target)

    def test_repr_v3(self):
        """Test method ``nailgun.entity_mixins.Entity.__repr__``.

        Assert that ``__repr__`` works correctly when one entity has a foreign
        key relationship to a second entity.

        """
        entity_id = gen_integer()
        target = (
            'tests.test_entity_mixins.SampleEntityTwo('
            '{0}, '
            'one_to_many=[tests.test_entity_mixins.SampleEntity({0}, id={1})]'
            ')'
            .format(self.cfg, entity_id)
        )
        entity = SampleEntityTwo(
            self.cfg,
            one_to_many=[SampleEntity(self.cfg, id=entity_id)]
        )
        self.assertEqual(repr(entity), target)
        import nailgun  # noqa pylint:disable=unused-variable
        import tests  # noqa pylint:disable=unused-variable
        # pylint:disable=eval-used
        self.assertEqual(repr(eval(repr(entity))), target)


class EntityCreateMixinTestCase(TestCase):
    """Tests for :class:`nailgun.entity_mixins.EntityCreateMixin`."""

    def setUp(self):
        """Set ``self.entity = EntityWithCreate(…)``."""
        self.entity = EntityWithCreate(
            config.ServerConfig('example.com'),
            id=gen_integer(min_value=1),
        )

    def test_create_missing(self):
        """Call method ``create_missing``."""

        class FKEntityWithCreate(
                entity_mixins.Entity,
                entity_mixins.EntityCreateMixin):
            """An entity that can be created and has foreign key fields."""

            def __init__(self, server_config=None, **kwargs):
                self._fields = {
                    'int': IntegerField(required=True),
                    'int_choices': IntegerField(choices=(1, 2), required=True),
                    'int_default': IntegerField(default=5, required=True),
                    'many': OneToManyField(SampleEntity, required=True),
                    'one': OneToOneField(SampleEntity, required=True),
                }
                super(FKEntityWithCreate, self).__init__(
                    server_config,
                    **kwargs
                )

        cfg = config.ServerConfig('example.com')
        entity = FKEntityWithCreate(cfg)
        with mock.patch.object(entity._fields['many'], 'gen_value') as gen1:
            with mock.patch.object(entity._fields['one'], 'gen_value') as gen2:
                self.assertEqual(entity.create_missing(), None)
        for gen_value in gen1, gen2:
            self.assertEqual(
                gen_value.mock_calls,
                [
                    mock.call(),  # gen_value() returns a class. The returned
                    mock.call()(cfg),  # class is instantiated, and
                    mock.call()().create(True),  # create(True) is called.
                ]
            )
        self.assertEqual(
            set(entity.get_fields().keys()) - {'id'},
            set(entity.get_values().keys()),
        )
        self.assertIn(entity.int_choices, (1, 2))  # pylint:disable=no-member
        self.assertEqual(entity.int_default, 5)  # pylint:disable=no-member

    def test_create_raw_v1(self):
        """What happens if the ``create_missing`` arg is not specified?

        :meth:`nailgun.entity_mixins.EntityCreateMixin.create_raw` should
        default to :data:`nailgun.entity_mixins.CREATE_MISSING`. We do not set
        ``CREATE_MISSING`` in this test. It is a process-wide variable, and
        setting it may prevent tests from being run in parallel.

        """
        with mock.patch.object(self.entity, 'create_missing') as c_missing:
            with mock.patch.object(self.entity, 'create_payload') as c_payload:
                with mock.patch.object(client, 'post') as post:
                    self.entity.create_raw()
        self.assertEqual(
            c_missing.call_count,
            1 if entity_mixins.CREATE_MISSING else 0
        )
        self.assertEqual(c_payload.call_count, 1)
        self.assertEqual(post.call_count, 1)

    def test_create_raw_v2(self):
        """What happens if the ``create_missing`` arg is ``True``?"""
        with mock.patch.object(self.entity, 'create_missing') as c_missing:
            with mock.patch.object(self.entity, 'create_payload') as c_payload:
                with mock.patch.object(client, 'post') as post:
                    self.entity.create_raw(True)
        self.assertEqual(c_missing.call_count, 1)
        self.assertEqual(c_payload.call_count, 1)
        self.assertEqual(post.call_count, 1)

    def test_create_raw_v3(self):
        """What happens if the ``create_missing`` arg is ``False``?"""
        with mock.patch.object(self.entity, 'create_missing') as c_missing:
            with mock.patch.object(self.entity, 'create_payload') as c_payload:
                with mock.patch.object(client, 'post') as post:
                    self.entity.create_raw(False)
        self.assertEqual(c_missing.call_count, 0)
        self.assertEqual(c_payload.call_count, 1)
        self.assertEqual(post.call_count, 1)

    def test_create_json(self):
        """Test :meth:`nailgun.entity_mixins.EntityCreateMixin.create_json`."""
        for create_missing in (None, True, False):
            response = mock.Mock()
            response.json.return_value = gen_integer()
            with mock.patch.object(self.entity, 'create_raw') as create_raw:
                create_raw.return_value = response
                self.entity.create_json(create_missing)
            self.assertEqual(create_raw.call_count, 1)
            self.assertEqual(create_raw.call_args[0][0], create_missing)
            self.assertEqual(response.raise_for_status.call_count, 1)
            self.assertEqual(response.json.call_count, 1)

    def test_create(self):
        """Test :meth:`nailgun.entity_mixins.EntityCreateMixin.create`."""

        class EntityWithCreateRead(
                EntityWithCreate,
                entity_mixins.EntityReadMixin):
            """An entity that can be created and read."""

        readable = EntityWithCreateRead(
            config.ServerConfig('example.com'),
            id=gen_integer(),
        )
        for create_missing in (None, True, False):
            with mock.patch.object(readable, 'create_json') as create_json:
                create_json.return_value = gen_integer()
                with mock.patch.object(readable, 'read') as read:
                    readable.create(create_missing)
            self.assertEqual(create_json.call_count, 1)
            self.assertEqual(create_json.call_args[0][0], create_missing)
            self.assertEqual(read.call_count, 1)
            self.assertEqual(
                read.call_args[1]['attrs'],
                create_json.return_value,
            )


class EntityReadMixinTestCase(TestCase):
    """Tests for :class:`nailgun.entity_mixins.EntityReadMixin`."""

    @classmethod
    def setUpClass(cls):
        """Set ``cls.test_entity``.

        ``test_entity`` is a class having one to one and one to many fields.

        """

        class TestEntity(entity_mixins.Entity, entity_mixins.EntityReadMixin):
            """An entity with several different types of fields."""

            def __init__(self, server_config=None, **kwargs):
                self._fields = {
                    'ignore_me': IntegerField(),
                    'many': OneToManyField(SampleEntity),
                    'none': OneToOneField(SampleEntity),
                    'one': OneToOneField(SampleEntity),
                }
                self._meta = {'api_path': ''}
                super(TestEntity, self).__init__(server_config, **kwargs)

        cls.test_entity = TestEntity

    def setUp(self):
        """Set ``self.entity = EntityWithRead(…)``."""
        self.cfg = config.ServerConfig('example.com')
        self.entity = EntityWithRead(self.cfg, id=gen_integer(min_value=1))

    def test_read_raw(self):
        """Call :meth:`nailgun.entity_mixins.EntityReadMixin.read_raw`."""
        with mock.patch.object(client, 'get') as get:
            self.entity.read_raw()
        self.assertEqual(get.call_count, 1)
        self.assertEqual(len(get.call_args[0]), 1)  # path='…'
        self.assertEqual(get.call_args[0][0], self.entity.path())
        self.assertEqual(
            get.call_args[1],
            # pylint:disable=protected-access
            self.entity._server_config.get_client_kwargs(),
        )

    def test_read_json(self):
        """Call :meth:`nailgun.entity_mixins.EntityReadMixin.read_json`."""
        response = mock.Mock()
        response.json.return_value = gen_integer()
        with mock.patch.object(self.entity, 'read_raw') as read_raw:
            read_raw.return_value = response
            self.entity.read_json()
        self.assertEqual(read_raw.call_count, 1)
        self.assertEqual(response.raise_for_status.call_count, 1)
        self.assertEqual(response.json.call_count, 1)

    def test_read_v1(self):
        """Make ``read_json`` return hashes."""
        # Generate some bogus values and call `read`.
        entity_1 = self.test_entity(self.cfg)
        attrs = {
            'id': gen_integer(min_value=1),
            'manies': [{'id': gen_integer(min_value=1)}],
            'none': None,
            'one': {'id': gen_integer(min_value=1)},
        }
        with mock.patch.object(entity_1, 'read_json') as read_json:
            read_json.return_value = attrs
            entity_2 = entity_1.read(ignore={'ignore_me'})

        # Make assertions about the call and the returned entity.
        self.assertEqual(
            entity_2._server_config,  # pylint:disable=protected-access
            self.cfg,
        )
        self.assertEqual(read_json.call_count, 1)
        self.assertEqual(
            set(entity_1.get_fields().keys()),
            set(entity_2.get_fields().keys()),
        )
        self.assertEqual(entity_2.id, attrs['id'])
        self.assertEqual(entity_2.many[0].id, attrs['manies'][0]['id'])
        self.assertEqual(entity_2.one.id, attrs['one']['id'])

    def test_read_v2(self):
        """Make ``read_json`` return hashes, but with different field names."""
        # Generate some bogus values and call `read`.
        entity_1 = self.test_entity(self.cfg)
        attrs = {'many': [{'id': gen_integer(min_value=1)}]}
        with mock.patch.object(entity_1, 'read_json') as read_json:
            read_json.return_value = attrs
            entity_2 = entity_1.read(ignore={'id', 'none', 'one', 'ignore_me'})

        # Make assertions about the call and the returned entity.
        self.assertEqual(
            entity_2._server_config,  # pylint:disable=protected-access
            self.cfg,
        )
        self.assertEqual(read_json.call_count, 1)
        self.assertEqual(
            set(entity_1.get_fields().keys()),
            set(entity_2.get_fields().keys()),
        )
        self.assertEqual(entity_2.many[0].id, attrs['many'][0]['id'])

    def test_read_v3(self):
        """Make ``read_json`` return IDs."""
        # Generate some bogus values and call `read`.
        entity_1 = self.test_entity(self.cfg)
        attrs = {
            'id': gen_integer(min_value=1),
            'many_ids': [gen_integer(min_value=1)],
            'none': None,
            'one_id': gen_integer(min_value=1),
        }
        with mock.patch.object(entity_1, 'read_json') as read_json:
            read_json.return_value = attrs
            entity_2 = entity_1.read(ignore={'ignore_me'})

        # Make assertions about the call and the returned entity.
        self.assertEqual(
            entity_2._server_config,  # pylint:disable=protected-access
            self.cfg,
        )
        self.assertEqual(read_json.call_count, 1)
        self.assertEqual(
            set(entity_1.get_fields().keys()),
            set(entity_2.get_fields().keys()),
        )
        self.assertEqual(entity_2.id, attrs['id'])
        self.assertEqual(entity_2.many[0].id, attrs['many_ids'][0])
        self.assertEqual(entity_2.one.id, attrs['one_id'])

    def test_read_v4(self):
        """Do not ignore any fields."""
        with mock.patch.object(
            entity_mixins.EntityReadMixin,
            'read_json',
            return_value={'id': gen_integer()},
        ) as read_json:
            entity = EntityWithRead(self.cfg).read()
        self.assertEqual(entity.get_values(), read_json.return_value)

    def test_missing_value_error(self):
        """Raise a :class:`nailgun.entity_mixins.MissingValueError`."""
        entity = self.test_entity(config.ServerConfig('example.com'))
        for attrs in (
                {
                    'id': gen_integer(min_value=1),
                    'none': None,
                    'one_id': gen_integer(min_value=1),
                },
                {
                    'id': gen_integer(min_value=1),
                    'many_ids': [gen_integer(min_value=1)],
                    'none': None,
                },
        ):
            with self.subTest(attrs):
                with mock.patch.object(entity, 'read_json') as read_json:
                    read_json.return_value = attrs
                    with self.assertRaises(entity_mixins.MissingValueError):
                        entity.read(ignore={'ignore_me'})


class EntityUpdateMixinTestCase(TestCase):
    """Tests for :class:`nailgun.entity_mixins.EntityUpdateMixin`."""

    def setUp(self):
        """Set ``self.entity = EntityWithUpdate(…)``."""
        self.entity = EntityWithUpdate(
            config.ServerConfig('example.com'),
            id=gen_integer(min_value=1),
        )

    def test_update_payload_v1(self):
        """Call :meth:`nailgun.entity_mixins.EntityUpdateMixin.update_payload`.

        Assert that the method behaves correctly given various values for the
        ``field`` argument.

        """

        class TestEntity(EntityWithUpdate):
            """Just like its parent class, but with fields."""

            def __init__(self, server_config=None, **kwargs):
                self._fields = {'one': IntegerField(), 'two': IntegerField()}
                super(TestEntity, self).__init__(server_config, **kwargs)

        cfg = config.ServerConfig('url')
        args_list = (
            {},
            {'one': gen_integer()},
            {'two': gen_integer()},
            {'one': gen_integer(), 'two': gen_integer()},
        )

        # Make `update_payload` return all or no values.
        for args in args_list:
            entity = TestEntity(cfg, **args)
            self.assertEqual(entity.update_payload(), args)
            self.assertEqual(entity.update_payload(list(args.keys())), args)
            self.assertEqual(entity.update_payload([]), {})

        # Make `update_payload` return only some values.
        entity = TestEntity(cfg, **args_list[-1])
        self.assertEqual(
            entity.update_payload(['one']),
            {'one': args_list[-1]['one']},
        )
        self.assertEqual(
            entity.update_payload(['two']),
            {'two': args_list[-1]['two']},
        )

        # Ask `update_payload` to return unavailable values.
        entity = TestEntity(cfg)
        for field_names in (['one'], ['two'], ['one', 'two']):
            with self.assertRaises(KeyError):
                entity.update_payload(field_names)

    def test_update_payload_v2(self):
        """Call :meth:`nailgun.entity_mixins.EntityUpdateMixin.update_payload`.

        Assign ``None`` to a ``OneToOneField`` and call ``update_payload``.

        """

        class TestEntity(EntityWithUpdate):
            """Just like its parent class, but with fields."""

            def __init__(self, server_config=None, **kwargs):
                self._fields = {'other': OneToOneField(SampleEntity)}
                super(TestEntity, self).__init__(server_config, **kwargs)

        cfg = config.ServerConfig('url')
        entities = [TestEntity(cfg, other=None), TestEntity(cfg)]
        entities[1].other = None  # pylint:disable=W0201
        for entity in entities:
            with self.subTest(entity):
                self.assertEqual(entity.update_payload(), {'other_id': None})

    def test_update_raw(self):
        """Call :meth:`nailgun.entity_mixins.EntityUpdateMixin.update_raw`."""
        with mock.patch.object(self.entity, 'update_payload') as u_payload:
            with mock.patch.object(client, 'put') as put:
                self.entity.update_raw()
        self.assertEqual(u_payload.call_count, 1)
        self.assertEqual(put.call_count, 1)
        self.assertEqual(len(put.call_args[0]), 2)  # path='…' and data={…}
        self.assertEqual(put.call_args[0][0], self.entity.path())
        self.assertEqual(
            put.call_args[1],
            # pylint:disable=protected-access
            self.entity._server_config.get_client_kwargs(),
        )

    def test_update_json(self):
        """Call :meth:`nailgun.entity_mixins.EntityUpdateMixin.update_json`."""
        response = mock.Mock()
        response.json.return_value = gen_integer()
        with mock.patch.object(self.entity, 'update_raw') as update_raw:
            update_raw.return_value = response
            self.entity.update_json()
        self.assertEqual(update_raw.call_count, 1)
        self.assertEqual(response.raise_for_status.call_count, 1)
        self.assertEqual(response.json.call_count, 1)

    def test_update(self):
        """Test :meth:`nailgun.entity_mixins.EntityUpdateMixin.update`."""

        class EntityWithUpdateRead(
                EntityWithUpdate,
                entity_mixins.EntityReadMixin):
            """An entity that can be updated and read."""

        readable = EntityWithUpdateRead(
            config.ServerConfig('example.com'),
            id=gen_integer(),
        )
        with mock.patch.object(readable, 'update_json') as update_json:
            update_json.return_value = gen_integer()
            with mock.patch.object(readable, 'read') as read:
                readable.update()
        self.assertEqual(update_json.call_count, 1)
        self.assertEqual(read.call_count, 1)
        self.assertEqual(
            read.call_args[1]['attrs'],
            update_json.return_value,
        )


class EntityDeleteMixinTestCase(TestCase):
    """Tests for :class:`nailgun.entity_mixins.EntityDeleteMixin`."""

    def setUp(self):
        """Set ``self.entity = EntityWithDelete(…)``."""
        self.entity = EntityWithDelete(
            config.ServerConfig('example.com'),
            id=gen_integer(min_value=1),
        )

    def test_delete_raw(self):
        """Call :meth:`nailgun.entity_mixins.EntityDeleteMixin.delete_raw`."""
        with mock.patch.object(client, 'delete') as delete:
            self.entity.delete_raw()
        self.assertEqual(delete.call_count, 1)
        self.assertEqual(len(delete.call_args[0]), 1)
        self.assertEqual(delete.call_args[0][0], self.entity.path())
        self.assertEqual(
            delete.call_args[1],
            # pylint:disable=protected-access
            self.entity._server_config.get_client_kwargs(),
        )

    def test_delete_v1(self):
        """What happens if the server returns an error HTTP status code?"""
        response = mock.Mock()
        response.raise_for_status.side_effect = HTTPError('oh no!')
        with mock.patch.object(
            entity_mixins.EntityDeleteMixin,
            'delete_raw',
            return_value=response,
        ):
            with self.assertRaises(HTTPError):
                self.entity.delete()

    def test_delete_v2(self):
        """What happens if the server returns an HTTP ACCEPTED status code?"""
        response = mock.Mock()
        response.status_code = http_client.ACCEPTED
        response.json.return_value = {'id': gen_integer()}
        with mock.patch.object(
            entity_mixins.EntityDeleteMixin,
            'delete_raw',
            return_value=response,
        ) as delete_raw:
            with mock.patch.object(entity_mixins, '_poll_task') as poll_task:
                self.entity.delete()
        self.assertEqual(delete_raw.call_count, 1)
        self.assertEqual(poll_task.call_count, 1)
        self.assertEqual(
            poll_task.call_args[0],  # a tuple of (positional, keyword) args
            # pylint:disable=protected-access
            (response.json.return_value['id'], self.entity._server_config)
        )

    def test_delete_v3(self):
        """What happens if the server returns an HTTP NO_CONTENT status?"""
        response = mock.Mock()
        response.status_code = http_client.NO_CONTENT
        with mock.patch.object(
            entity_mixins.EntityDeleteMixin,
            'delete_raw',
            return_value=response,
        ):
            with mock.patch.object(entity_mixins, '_poll_task') as poll_task:
                self.assertEqual(self.entity.delete(), None)
        self.assertEqual(poll_task.call_count, 0)

    def test_delete_v4(self):
        """What happens if the server returns some other stuccess status?"""
        response = mock.Mock()
        response.json.return_value = gen_integer()
        with mock.patch.object(
            entity_mixins.EntityDeleteMixin,
            'delete_raw',
            return_value=response,
        ):
            self.assertEqual(self.entity.delete(), response.json.return_value)


class EntitySearchMixinTestCase(TestCase):
    """Tests for :class:`nailgun.entity_mixins.EntitySearchMixin`."""

    def setUp(self):
        """Set ``self.cfg`` and ``self.entity``."""
        self.cfg = config.ServerConfig('example.com')
        self.entity = EntityWithSearch(self.cfg)

    def test_search_payload_v1(self):
        """Call ``search_payload``. Generate an empty query."""
        for kwargs in (
                {'fields': set()},
                {'query': {}, 'fields': set()},
                {'query': {}},
                {},
        ):
            with self.subTest(kwargs):
                self.assertEqual(self.entity.search_payload(**kwargs), {})

    def test_search_payload_v2(self):
        """Call ``search_payload``. Pass in a query."""
        query = {'foo': 'bar'}
        self.assertEqual(self.entity.search_payload([], query), query)

    def test_search_payload_v3(self):
        """Call ``search_payload``. Include a variety of fields in a search."""
        inputs_outputs = (
            ({'id': 1}, {'id': 1}),
            ({'one': 2}, {'one_id': 2}),
            ({'many': [3, 4]}, {'many_ids': [3, 4]}),
        )
        entity = EntityWithSearch2(self.cfg, id=1, one=2, many=[3, 4])

        # Implicitly and explicitly select one field for the payload.
        for input_, output in inputs_outputs:
            self.assertEqual(
                EntityWithSearch2(self.cfg, **input_).search_payload(),
                output,
            )
        for input_, output in inputs_outputs:
            self.assertEqual(
                entity.search_payload(set(input_.keys())),
                output,
            )

        # Implicitly and explicitly include all fields in the payload.
        self.assertEqual(
            entity.search_payload(),
            {'id': 1, 'one_id': 2, 'many_ids': [3, 4]},
        )
        self.assertEqual(
            entity.search_payload({'id', 'one', 'many'}),
            {'id': 1, 'one_id': 2, 'many_ids': [3, 4]},
        )

    def test_search_raw(self):
        """Call :meth:`nailgun.entity_mixins.EntitySearchMixin.search_raw`."""
        with mock.patch.object(self.entity, 'search_payload') as s_payload:
            with mock.patch.object(client, 'get') as get:
                self.entity.search_raw()
        self.assertEqual(s_payload.call_count, 1)
        self.assertEqual(get.call_count, 1)
        self.assertEqual(len(get.call_args[0]), 1)  # path='…'
        self.assertEqual(get.call_args[0][0], self.entity.path())
        self.assertEqual(get.call_args[1]['data'], s_payload.return_value)

    def test_search_json(self):
        """Call :meth:`nailgun.entity_mixins.EntitySearchMixin.search_json`."""
        response = mock.Mock()
        response.json.return_value = gen_integer()
        with mock.patch.object(self.entity, 'search_raw') as search_raw:
            search_raw.return_value = response
            self.entity.search_json()
        self.assertEqual(search_raw.call_count, 1)
        self.assertEqual(response.raise_for_status.call_count, 1)
        self.assertEqual(response.json.call_count, 1)

    def test_search_normalize_v1(self):
        """Call ``search_normalize``.

        Pretend the server returns values for all fields, and an extra value.

        """
        with mock.patch.object(entity_mixins, '_get_entity_ids') as get_ids:
            with mock.patch.object(entity_mixins, '_get_entity_id') as get_id:
                attrs_list = EntityWithSearch2(self.cfg).search_normalize([{
                    'extra': 'foo',  # simulate extra value returned by server
                    'id': 'bar',
                    'many_ids': [gen_integer()],
                    'one_id': gen_integer(),
                }])
        self.assertEqual(get_ids.call_count, 1)
        self.assertEqual(get_id.call_count, 1)
        self.assertEqual(len(attrs_list), 1)
        self.assertEqual(
            attrs_list[0],
            {
                'id': 'bar',
                'many': get_ids.return_value,
                'one': get_id.return_value,
            }
        )

    def test_search_normalize_v2(self):
        """Call ``search_normalize``.

        Pretend the server returns no values for any fields.

        """
        attrs_list = EntityWithSearch2(self.cfg).search_normalize([{}])
        self.assertEqual(len(attrs_list), 1)
        self.assertEqual(attrs_list[0], {})

    def test_search_v1(self):
        """Test :meth:`nailgun.entity_mixins.EntitySearchMixin.search`.

        Pass no arguments.

        """
        with mock.patch.object(self.entity, 'search_normalize') as s_normalize:
            with mock.patch.object(self.entity, 'search_filter') as s_filter:
                with mock.patch.object(self.entity, 'search_json') as s_json:
                    results = self.entity.search()
        self.assertEqual(s_json.call_count, 1)
        self.assertEqual(s_normalize.call_count, 1)
        self.assertEqual(s_filter.call_count, 0)
        self.assertEqual(results, [])

    def test_search_v2(self):
        """Test :meth:`nailgun.entity_mixins.EntitySearchMixin.search`.

        Provide each possible argument.

        """
        with mock.patch.object(self.entity, 'search_normalize') as s_normalize:
            with mock.patch.object(self.entity, 'search_filter') as s_filter:
                with mock.patch.object(self.entity, 'search_json') as s_json:
                    s_normalize.return_value = [{'id': 'foo'}]
                    s_filter.return_value = [gen_integer()]  # not realistic
                    results = self.entity.search('fields', 'query', 'filters')
        self.assertEqual(s_json.call_count, 1)
        self.assertEqual(s_normalize.call_count, 1)
        self.assertEqual(s_filter.call_count, 1)
        self.assertEqual(results, s_filter.return_value)

        # Check search_json
        self.assertEqual(len(s_json.call_args[0]), 2)  # fields, query
        self.assertEqual(len(s_json.call_args[1]), 0)
        self.assertEqual(s_json.call_args[0][0], 'fields')
        self.assertEqual(s_json.call_args[0][1], 'query')

        # Check search_normalize
        self.assertEqual(len(s_normalize.call_args[0]), 1)  # results
        self.assertEqual(len(s_normalize.call_args[1]), 0)
        self.assertEqual(
            s_normalize.call_args[0][0],
            s_json.return_value['results'],
        )

        # Check search_filter
        self.assertEqual(len(s_filter.call_args[0]), 2)  # entities, filters
        self.assertEqual(len(s_filter.call_args[1]), 0)
        self.assertEqual(len(s_filter.call_args[0][0]), 1)
        self.assertEqual(s_filter.call_args[0][0][0].id, 'foo')  # from mock ↑
        self.assertEqual(s_filter.call_args[0][1], 'filters')

    def test_search_filter_v1(self):
        """Test :meth:`nailgun.entity_mixins.EntitySearchMixin.search_filter`.

        Pass a zero-length list of entities.

        """
        self.assertEqual(
            [],
            entity_mixins.EntitySearchMixin.search_filter([], {}),
        )

    def test_search_filter_v2(self):
        """Test :meth:`nailgun.entity_mixins.EntitySearchMixin.search_filter`.

        Try to filter on a foreign key field.

        """
        for filter_ in ({'one': 'foo'}, {'many': 'bar'}):
            with self.subTest(filter_):
                with self.assertRaises(NotImplementedError):
                    entity_mixins.EntitySearchMixin.search_filter(
                        [EntityWithSearch2(self.cfg)],
                        filter_,
                    )

    def test_search_filter_v3(self):
        """Test :meth:`nailgun.entity_mixins.EntitySearchMixin.search_filter`.

        Pass an invalid filter.

        """
        with self.assertRaises(entity_mixins.NoSuchFieldError):
            entity_mixins.EntitySearchMixin.search_filter(
                [EntityWithSearch2(self.cfg)],
                {'field name': 'field value'},
            )

    def test_search_filter_v4(self):
        """Test :meth:`nailgun.entity_mixins.EntitySearchMixin.search_filter`.

        Pass in valid entities and filters.

        """

        class EntityWithSearch3(
                EntityWithSearch, entity_mixins.EntityReadMixin):
            """An entity inheriting from the search and read mixins."""

        with mock.patch.object(EntityWithSearch3, 'read') as read:
            read.return_value = EntityWithSearch3(self.cfg, id=15)
            results = entity_mixins.EntitySearchMixin.search_filter(
                [EntityWithSearch3(self.cfg)],
                {'id': 15},
            )
        self.assertEqual(read.call_count, 1)
        self.assertEqual(results, [read.return_value])



# coding: utf-8
"""Tests for :mod:`nailgun.signals`."""
# pylint: disable-all
from sys import version_info
from mock import patch
from nailgun import entities, config, signals, client

if version_info < (3, 4):
    from unittest2 import TestCase
else:
    from unittest import TestCase

ORG_DATA = {'id': 1, 'domain': 'http://example.com'}


class FakeResponse(object):
    """Faking here cause we only need to test signals"""
    status_code = 200

    def __init__(self, data=None):
        self.data = data or ORG_DATA

    def raise_for_status(self):
        pass

    def json(self):
        return self.data


class Organization(entities.Organization):
    """Faking Organization as example to ensure nothing will be performed
    except the signals"""

    def read(self, *args, **kwargs):
        return self

    def create_json(self, create_missing=None):
        return self.__dict__

    def update_json(self, fields=None):
        pass

    def delete_raw(self):
        return FakeResponse()

    def _get_entity_ids(self):
        return {}

    def search_json(self, fields=None, query=None):
        return {"results": [ORG_DATA]}

    def search_normalize(self, results):
        return [ORG_DATA]

    @staticmethod
    def search_filter(entities, filters):
        return entities


cfg = config.ServerConfig(ORG_DATA['domain'])
org = Organization(cfg, **ORG_DATA)


@patch('nailgun.entities.Organization.read', return_value=org)
@patch('nailgun.client.get', return_value=FakeResponse())
class SignalsTestCase(TestCase):
    """This test case tests only if all signals are emitted
    and if the arguments were passed and catch by connected listeners
    """

    def test_signals_available(self, *args, **kwargs):
        self.assertTrue(signals.SIGNALS_AVAILABLE)

    def test_patching(self, *args, **kwargs):
        '''Make sure patching works'''
        self.assertIsInstance(client.get('http://example.com'), FakeResponse)
        self.assertEqual(entities.Organization(cfg).read(), org)

    def test_pre_create_signal(self, *args, **kwargs):
        self.assertIsNone(getattr(org, 'pre_create_emitted', None))

        @signals.pre_create.connect
        def pre_create(sender, create_missing, **kwargs):
            sender.pre_create_emitted = True

        org.create()
        self.assertTrue(org.pre_create_emitted)

    def test_post_create_signal(self, *args, **kwargs):
        self.assertIsNone(getattr(org, 'post_create_emitted', None))

        @signals.post_create.connect
        def post_create(sender, entity, **kwargs):
            entity.post_create_emitted = True

        org.create()
        self.assertTrue(org.post_create_emitted)

    def test_pre_update_signal(self, *args, **kwargs):
        self.assertIsNone(getattr(org, 'pre_update_emitted', None))

        @signals.pre_update.connect
        def pre_update(sender, fields, **kwargs):
            sender.pre_update_emitted = True

        org.update()
        self.assertTrue(org.pre_update_emitted)

    def test_post_update_signal(self, *args, **kwargs):
        self.assertIsNone(getattr(org, 'post_update_emitted', None))

        @signals.post_update.connect
        def post_update(sender, entity, fields, **kwargs):
            entity.post_update_emitted = True

        org.update()
        self.assertTrue(org.post_update_emitted)

    def test_pre_delete_signal(self, *args, **kwargs):
        self.assertIsNone(getattr(org, 'pre_delete_emitted', None))

        @signals.pre_delete.connect
        def pre_delete(sender, synchronous, **kwargs):
            sender.pre_delete_emitted = True

        org.delete()
        self.assertTrue(org.pre_delete_emitted)

    def test_post_delete_signal(self, *args, **kwargs):
        self.assertIsNone(getattr(org, 'post_delete_emitted', None))

        @signals.post_delete.connect
        def post_delete(sender, synchronous, result, **kwargs):
            sender.post_delete_emitted = True

        deletion_result = org.delete()
        self.assertTrue(org.post_delete_emitted)
        self.assertEqual(deletion_result, ORG_DATA)

    def test_pre_search_signal(self, *args, **kwargs):
        self.assertIsNone(getattr(org, 'pre_search_emitted', None))

        @signals.pre_search.connect
        def pre_search(sender, fields, query, filters, **kwargs):
            sender.pre_search_emitted = True

        org.search()
        self.assertTrue(org.pre_search_emitted)

    def test_post_search_signal(self, *args, **kwargs):
        self.assertIsNone(getattr(org, 'post_search_emitted', None))

        @signals.post_search.connect
        def post_search(sender, entities, fields, query, filters, **kwargs):
            sender.post_search_emitted = True

        org.search()
        self.assertTrue(org.post_search_emitted)



# -*- coding: utf-8 -*-
"""Defines a set of named signals for entities operations"""
# pylint: disable-all

import blinker_herald

emit = blinker_herald.emit
SENDER_CLASS = blinker_herald.SENDER_CLASS
SIGNALS_AVAILABLE = blinker_herald.SIGNALS_AVAILABLE

pre_create = blinker_herald.signals_namespace.signal('pre_create')
post_create = blinker_herald.signals_namespace.signal('post_create')

pre_update = blinker_herald.signals_namespace.signal('pre_update')
post_update = blinker_herald.signals_namespace.signal('post_update')

pre_delete = blinker_herald.signals_namespace.signal('pre_delete')
post_delete = blinker_herald.signals_namespace.signal('post_delete')

pre_search = blinker_herald.signals_namespace.signal('pre_search')
post_search = blinker_herald.signals_namespace.signal('post_search')



"""Unit tests for :mod:`nailgun.config`."""
from mock import call, mock_open, patch
from nailgun.config import (
    BaseServerConfig, ServerConfig, _get_config_file_path, ConfigFileError
)
from packaging.version import parse
import json

from sys import version_info
if version_info.major == 2:
    # The `__builtins__` module (note the "s") also provides the `open`
    # function. However, that module is an implementation detail for CPython 2,
    # so it should not be relied on.
    import __builtin__ as builtins  # pylint:disable=import-error
else:
    import builtins  # pylint:disable=import-error
if version_info < (3, 4):
    from unittest2 import TestCase  # pylint:disable=import-error
else:
    from unittest import TestCase


FILE_PATH = '/tmp/bogus.json'
CONFIGS = {
    'default': {'url': 'http://example.com'},
    'Ask Aak': {'url': 'bogus value', 'auth': ['username', 'password']},
    'Acros-Krik': {'url': 'booger', 'version': '1.2.3.4.dev5+67'},
    'King Adas': {'url': 'burger', 'version': 'silly version'},
}
CONFIGS2 = CONFIGS.copy()
CONFIGS2.update({
    'Abeloth': {'url': 'bogus value', 'verify': True},
    'Admiral Gial Ackbar': {'url': 'bogus', 'auth': [], 'verify': False},
})


def _convert_bsc_attrs(bsc_attrs):
    """Alter a dict of attributes in the same way as ``BaseServerConfig``.

    ``bsc_attrs`` is not altered. If changes are made, a copy is returned.

    :param bsc_attrs: A dict of attributes as might be passed to
        :class:`nailgun.config.BaseServerConfig`.
    :returns: A dict of attributes as is returned by
        ``vars(BaseServerConfig(bsc_attrs))``.

    """
    if 'version' in bsc_attrs:
        bsc_attrs = bsc_attrs.copy()  # shadow the passed in dict
        bsc_attrs['version'] = parse(bsc_attrs.pop('version'))
    return bsc_attrs


class BaseServerConfigTestCase(TestCase):
    """Tests for :class:`nailgun.config.BaseServerConfig`."""

    def test_init(self):
        """Test instantiating :class:`nailgun.config.BaseServerConfig`.

        Assert that only provided values become object attributes.

        """
        for config in CONFIGS.values():
            self.assertEqual(
                _convert_bsc_attrs(config),
                vars(BaseServerConfig(**config)),
            )

    def test_get(self):
        """Test :meth:`nailgun.config.BaseServerConfig.get`.

        Assert that the method extracts the asked-for section from a
        configuration file and correctly populates a new ``BaseServerConfig``
        object. Also assert that the ``auth`` attribute is a list. (See the
        docstring for :meth:`nailgun.config.ServerConfig.get`.)

        """
        for label, config in CONFIGS.items():
            open_ = mock_open(read_data=json.dumps(CONFIGS))
            with patch.object(builtins, 'open', open_):
                server_config = BaseServerConfig.get(label, FILE_PATH)
            self.assertEqual(type(server_config), BaseServerConfig)
            open_.assert_called_once_with(FILE_PATH)
            self.assertEqual(_convert_bsc_attrs(config), vars(server_config))
            if hasattr(server_config, 'auth'):
                self.assertIsInstance(server_config.auth, list)

    def test_get_labels(self):
        """Test :meth:`nailgun.config.BaseServerConfig.get_labels`.

        Assert that the method returns the correct labels.

        """
        open_ = mock_open(read_data=json.dumps(CONFIGS))
        with patch.object(builtins, 'open', open_):
            self.assertEqual(
                set(CONFIGS.keys()),
                set(BaseServerConfig.get_labels(FILE_PATH)),
            )
        open_.assert_called_once_with(FILE_PATH)

    def test_save(self):
        """Test :meth:`nailgun.config.BaseServerConfig.save`.

        Assert that the method reads the config file before writing, and that
        it writes out a correct config file.

        """
        label = 'Ask Aak'
        config = {label: {'url': 'https://example.org'}}
        open_ = mock_open(read_data=json.dumps(CONFIGS))
        with patch.object(builtins, 'open', open_):
            BaseServerConfig(config[label]['url']).save(label, FILE_PATH)

        # We care about two things: that this method reads the config file
        # before writing, and that the written string is correct. The first is
        # easy to verify...
        self.assertEqual(
            open_.call_args_list,
            [call(FILE_PATH), call(FILE_PATH, 'w')],
        )
        # ...and the second is a PITA to verify.
        actual_config = _get_written_json(open_)
        target_config = CONFIGS.copy()
        target_config[label] = config[label]
        self.assertEqual(target_config, actual_config)

    def test_delete(self):
        """Test :meth:`nailgun.config.BaseServerConfig.delete`.

        Assert that the method reads the config file before writing, and that
        it writes out a correct config file.

        """
        open_ = mock_open(read_data=json.dumps(CONFIGS))
        with patch.object(builtins, 'open', open_):
            BaseServerConfig.delete('Ask Aak', FILE_PATH)

        # See `test_save` for further commentary on what's being done here.
        self.assertEqual(
            open_.call_args_list,
            [call(FILE_PATH), call(FILE_PATH, 'w')],
        )
        actual_config = _get_written_json(open_)
        target_config = CONFIGS.copy()
        del target_config['Ask Aak']
        self.assertEqual(target_config, actual_config)


class ServerConfigTestCase(TestCase):
    """Tests for :class:`nailgun.config.ServerConfig`."""

    def test_init(self):
        """Test instantiating :class:`nailgun.config.ServerConfig`.

        Assert that only provided values become object attributes.

        """
        for config in CONFIGS2.values():
            self.assertEqual(
                _convert_bsc_attrs(config),
                vars(ServerConfig(**config)),
            )

    def test_raise_config_file_error(self):
        """Should raise error if path not found"""
        with self.assertRaises(ConfigFileError):
            # pylint:disable=protected-access
            _get_config_file_path('foo', 'bar')

    def test_get_client_kwargs(self):
        """Test :meth:`nailgun.config.ServerConfig.get_client_kwargs`.

        Assert that:

        * ``get_client_kwargs`` returns all of the instance attributes from its
          object except the "url" attribute, and
        * no instance attributes from the object are removed.

        """
        for config in CONFIGS2.values():
            target = config.copy()
            target.pop('url')
            target.pop('version', None)
            server_config = ServerConfig(**config)
            self.assertDictEqual(target, server_config.get_client_kwargs())
            self.assertDictEqual(
                vars(ServerConfig(**config)),
                vars(server_config)
            )

    def test_get(self):
        """Test :meth:`nailgun.config.ServerConfig.get`.

        Assert that the ``auth`` attribute is a tuple.

        """
        for label in CONFIGS.keys():
            open_ = mock_open(read_data=json.dumps(CONFIGS))
            with patch.object(builtins, 'open', open_):
                server_config = ServerConfig.get(label, FILE_PATH)
            self.assertEqual(type(server_config), ServerConfig)
            if hasattr(server_config, 'auth'):
                self.assertIsInstance(server_config.auth, tuple)


class ReprTestCase(TestCase):
    """Test method ``nailgun.config.BaseServerConfig.__repr__``."""

    def test_bsc_v1(self):
        """Test :class:`nailgun.config.BaseServerConfig`.

        Assert that ``__repr__`` works correctly when ``url`` is specified.

        """
        target = "nailgun.config.BaseServerConfig(url='bogus')"
        self.assertEqual(target, repr(BaseServerConfig('bogus')))
        import nailgun  # noqa pylint:disable=unused-variable
        # pylint:disable=eval-used
        self.assertEqual(target, repr(eval(repr(BaseServerConfig('bogus')))))

    def test_bsc_v2(self):
        """Test :class:`nailgun.config.BaseServerConfig`.

        Assert that ``__repr__`` works correctly when ``url`` and ``auth`` are
        specified.

        """
        targets = (
            "nailgun.config.BaseServerConfig(url='flim', auth='flam')",
            "nailgun.config.BaseServerConfig(auth='flam', url='flim')",
        )
        self.assertIn(repr(BaseServerConfig('flim', auth='flam')), targets)
        import nailgun  # noqa pylint:disable=unused-variable
        # pylint:disable=eval-used
        self.assertIn(
            repr(eval(repr(BaseServerConfig('flim', auth='flam')))),
            targets
        )

    def test_bsc_v3(self):
        """Test :class:`nailgun.config.BaseServerConfig`.

        Assert that ``__repr__`` works correctly when ``url`` and ``version``
        are specified.

        """
        targets = (
            "nailgun.config.BaseServerConfig(url='flim', version='1')",
            "nailgun.config.BaseServerConfig(version='1', url='flim')",
        )
        self.assertIn(repr(BaseServerConfig('flim', version='1')), targets)

    def test_sc_v1(self):
        """Test :class:`nailgun.config.ServerConfig`.

        Assert that ``__repr__`` works correctly when only a URL is passed in.

        """
        target = "nailgun.config.ServerConfig(url='bogus')"
        self.assertEqual(target, repr(ServerConfig('bogus')))
        import nailgun  # noqa pylint:disable=unused-variable
        # pylint:disable=eval-used
        self.assertEqual(target, repr(eval(repr(ServerConfig('bogus')))))

    def test_sc_v2(self):
        """Test :class:`nailgun.config.ServerConfig`.

        Assert that ``__repr__`` works correctly when ``url`` and ``auth`` are
        specified.

        """
        targets = (
            "nailgun.config.ServerConfig(url='flim', auth='flam')",
            "nailgun.config.ServerConfig(auth='flam', url='flim')",
        )
        self.assertIn(repr(ServerConfig('flim', auth='flam')), targets)
        import nailgun  # noqa pylint:disable=unused-variable
        # pylint:disable=eval-used
        self.assertIn(
            repr(eval(repr(ServerConfig('flim', auth='flam')))),
            targets
        )

    def test_sc_v3(self):
        """Test :class:`nailgun.config.ServerConfig`.

        Assert that ``__repr__`` works correctly when ``url`` and ``version``
        are specified.

        """
        targets = (
            "nailgun.config.ServerConfig(url='flim', version='1')",
            "nailgun.config.ServerConfig(version='1', url='flim')",
        )
        self.assertIn(repr(ServerConfig('flim', version='1')), targets)

    def test_sc_v4(self):
        """Test :class:`nailgun.config.ServerConfig`.

        Assert that ``__repr__`` works correctly when ``url`` and ``verify``
        are specified.

        """
        targets = (
            "nailgun.config.ServerConfig(url='flim', verify='flub')",
            "nailgun.config.ServerConfig(verify='flub', url='flim')",
        )
        self.assertIn(repr(ServerConfig('flim', verify='flub')), targets)
        import nailgun  # noqa pylint:disable=unused-variable
        # pylint:disable=eval-used
        self.assertIn(
            repr(eval(repr(ServerConfig('flim', verify='flub')))),
            targets
        )


def _get_written_json(mock_obj):
    """Return the JSON that has been written to a mock `open` object."""
    # json.dump() calls write() for each individual JSON token.
    return json.loads(''.join(
        tuple(call_obj)[1][0]
        for call_obj
        in mock_obj().write.mock_calls
    ))
